
/*	ZX Spectrum Keyboard Handler

***	Copyright:	1993-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.Jul.94:	Started work on this file							KIO !
	11.Jun.95:	revised 'game' handling, added Kempston support		KIO !
	28.Oct.96:	redefined keyboard matrix as union{}					KIO !
*/
#include	<string.h>

#include	"kio.h"
#include	"keyboard.h"
#include	"preferences.h"
#include	"application.h"
#include	"menue.h"
#include	"display.h"
#include	"z80 main.h"
#include	"help+about.h"


// -----	used method: iBasicKeys or iGameKeys
Char		kbdMethod;

// -----	key translation table: macintosh keycode -> depressed specci keys & joystick port bits
kbd		zx_key[0x80];

// -----	Current state of specci keyboard matrix -----------------------------------------
kbd		keyboard;


// ======================================================================

/*	coding of keyboard translation table initialization tables:
		0x80 = add caps key
		0x08 = add symbol key
		0x00-0x70 = key's row
		0x00-0x04 = key's column
		0xFF = not assigned

	initialization table: game keyboard mapping
	Specci keys are generated from: (asuming US keyboard)
		1...0			1234567890
		q...p			qwert yuiop
		a...enter		asdfg hjkl;
		caps...space	`zxcv bnm,.
	additional the following key bindings:
		esc			caps + symbol shift
		shift			caps shift
		control		symbol shift
		alt			symbol shift
		enter		enter
		space		space
		delete		caps + 0
		cursors		caps + 5 ... caps + 8
		numberpad	fully translated except num lock key
*/
Char	gk[0x80] = 
{
	0x10,0x11,0x12,0x13,0x64,0x14,0x01,0x02,	// 0x00-0x07: a   s   d   f   h   g   z   x
	0x03,0x04,0xFF,0x74,0x20,0x21,0x22,0x23,	// 0x08-0x0F: c   v      b   q   w   e   r
	0x54,0x24,0x30,0x31,0x32,0x33,0x44,0x34,	// 0x10-0x17: y   t   1   2   3   4   6   5
	0xFF,0x41,0x43,0xFF,0x42,0x40,0xFF,0x51,	// 0x18-0x1F:    9   7      8   0(null)      O(oh)
	0x53,0xFF,0x52,0x50,0x60,0x61,0x63,0xFF,	// 0x20-0x27: u      i   p   enter   l   j   
	0x62,0x60,0xFF,0x71,0xFF,0x73,0x72,0x70,	// 0x28-0x2F: k   ';'      ','      n   m   '.'
	0xFF,0x70,0x00,0xC0,0xFF,0x08,0xFF,0xFF,	// 0x30-0x37:    space   `   delete      esc      
	0x00,0xFF,0x71,0x71,0xFF,0xFF,0xFF,0xFF,		// 0x38-0x3F: shift      option   control            
	0xFF,0x7A,0xFF,0x7C,0xFF,0x6A,0xFF,0xFF,	// 0x40-0x47:    [.]      [*]      [+]      
	0xFF,0xFF,0xFF,0x0C,0x60,0xFF,0x6B,0xFF,		// 0x48-0x4F:          [/]   [enter]      [-]   
	0xFF,0x69,0x40,0x30,0x31,0x32,0x33,0x34,	// 0x50-0x57:    [=]   [0]   [1]   [2]   [3]   [4]   [5]
	0x44,0x43,0xFF,0x42,0x41,0xFF,0xFF,0xFF,	// 0x58-0x5F: [6]   [7]      [8]   [9]         
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,		// 0x60-0x67:
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,		// 0x68-0x6F:
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,		// 0x70-0x77:
	0xFF,0xFF,0xFF,0xB4,0xC2,0xC4,0xC3,0xFF		// 0x78-0x7F:          left   right   down   up   
};


// -----	initialize zx_key[] mac->specci key translation table from gk[]
InitKeyboard(int method)
{	Char q; kbd *z;
	int i;

	kbdMethod = method;

	memset(zx_key,0xFF,0x80*sizeof(kbd));			// clear all

// init game keyboard mapping
	for ( i=0; i<0x80; i++ )
	{	q=gk[i];
		if (q!=0xff)
		{	z=zx_key+i;
			if (q&0x80)	{ q-=0x80; z->byte[0] &= ~0x01; }	// caps shift
			if (q&0x08)	{ q-=0x08; z->byte[7] &= ~0x02; }	// symbol shift
			z->byte[q>>4] &= ~(1<<(q&0x07));
		}
	}

// initialize kempston interface (,,,fire,up,down,left,right)
	zx_key[0x79].joy.kempston = ~0x01;
	zx_key[0x75].joy.kempston = ~0x02;
	zx_key[0x77].joy.kempston = ~0x04;
	zx_key[0x73].joy.kempston = ~0x08;
	zx_key[0x74].joy.kempston = ~0x10;

}


// -----	Tastaturbersetzung fr Spiele:
//		Mehrere Tasten gleichzeitig
//		Tastatur im Gamemode entsprechend der Tastenlage
//		Umschalttaste  =  CAPS SHIFT
//		Option Taste  =  SYMBOL SHIFT
//		Es werden nur die Hauptbezeichner erkannt: 
//		"0" - "9", "A" - "Z", CAPS SHIFT, SYMBOL SHIFT, SPACE und ENTER
//		Cursortasten  =  CAPS SHIFT + "5" - "8"
//		ESC  =  CAPS SHIFT + SYMBOL SHIFT
//
//		Es gibt nur eine gemeinsame Prozedur fr HandleKEYUP und HandleKEYDOWN

UpdateKeyboard()
{	Char map[16];
	Char *p;
	int i,n,c;

	if ( (*(long*)kbdpict & *(long*)(kbdpict+4)) != -1 )
	{	if (!Button())
		{	static short cnt=1;
			if (cnt--) return; else cnt=1;
			if (helpscreen)
			{	SetPort(helpscreen);
				InvalRect(&btnbox);
			}
			*(long*)kbdpict = -1;
			*(long*)(kbdpict+4) = -1;
		}
	}	

	if (kbdMethod==iBasicKeys && mbar_present)  return;

	keyboard.quad.a = keyboard.quad.b = keyboard.quad.c = -1;

	GetKeys((Long*)map);

	if (map[6]&0x80)
	{	if (!mbar_present) MBarControl(show_mbar);
		return;
	}
	if (blacked_out&&mbar_present) 
	{	static int cnt=10; 
		if (!--cnt) { MBarControl(hide_mbar); cnt=10; }
	}

	p = map+16;
	for ( i=16; i--; )						// loop over bytes
	{	if (!*--p) continue;
		for ( c=*p, n=i*8;  c;  c>>=1, n++ )		// loop over bits
		{	if (c&1) 
			{	keyboard.quad.a &= zx_key[n].quad.a;	// left half of keyboard
				keyboard.quad.b &= zx_key[n].quad.b;	// right half of keyboard
				keyboard.quad.c &= zx_key[n].quad.c;	// joysticks
			};
		}
	}
}



// -----	Tastaturbersetzung Macintosh -> ZX Spectrum fr Basic ------------------------------------------

HandleKEYDOWN(Char key) 
{
	if (!z80_running) HandleFileMenu(iResume);	// restart z80 / set movie to record
	
	if (kbdMethod!=iBasicKeys) return;

	if ( key>='A' && key<='Z' )  {  keyboard.byte[0] &= 0xfe;  key+=0x20;  };
	
	switch ( (char)key )
	{
	case 27:	keyboard.byte[7] &= 0xfd;  keyboard.byte[0] &= 0xfe;  break;	// ESC -> extended command mode

// Token mittels Symbol Shift + Taste:

	case 'q'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[2] &= 0xfe;  break;	// <=
	case 'w'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[2] &= 0xfd;  break;	// <>
	case 'e'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[2] &= 0xfb;  break;	// >=
	case 'z'&31:	// for my convenience (German keyboard)
	case 'y'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[5] &= 0xef;  break;	// and
	case 'u'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[5] &= 0xf7;  break;	// or
	case 'i'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[5] &= 0xfb;  break;	// at

	case 'a'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[1] &= 0xfe;  break;	// STOP
	case 's'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[1] &= 0xfd;  break;	// NOT
	case 'd'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[1] &= 0xfb;  break;	// STEP
	case 'f'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[1] &= 0xf7;  break;	// TO	
	case 'g'&31:
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[1] &= 0xef;  break;	// THEN

// Einzelzeichen mittels Symbol Shift + Taste:
	case '.':	keyboard.byte[7] &= 0xfd;  keyboard.byte[7] &= 0xfb;  break;
	case ',':	keyboard.byte[7] &= 0xfd;  keyboard.byte[7] &= 0xf7;  break;	
	case '*':	keyboard.byte[7] &= 0xfd;  keyboard.byte[7] &= 0xef;  break;
	case '/':	keyboard.byte[7] &= 0xfd;  keyboard.byte[0] &= 0xef;  break;
	case '?':	keyboard.byte[7] &= 0xfd;  keyboard.byte[0] &= 0xf7;  break;	
	case '':	keyboard.byte[7] &= 0xfd;  keyboard.byte[0] &= 0xfb;  break;
	case ':':	keyboard.byte[7] &= 0xfd;  keyboard.byte[0] &= 0xfd;  break;
	case '=':	keyboard.byte[7] &= 0xfd;  keyboard.byte[6] &= 0xfd;  break;
	case '+':	keyboard.byte[7] &= 0xfd;  keyboard.byte[6] &= 0xfb;  break;
	case '-':	keyboard.byte[7] &= 0xfd;  keyboard.byte[6] &= 0xf7;  break;	
	case '^':	keyboard.byte[7] &= 0xfd;  keyboard.byte[6] &= 0xef;  break;
	case '"':	keyboard.byte[7] &= 0xfd;  keyboard.byte[5] &= 0xfe;  break;
	case ';':	keyboard.byte[7] &= 0xfd;  keyboard.byte[5] &= 0xfd;  break;
	case '>':	keyboard.byte[7] &= 0xfd;  keyboard.byte[2] &= 0xef;  break;
	case '<':	keyboard.byte[7] &= 0xfd;  keyboard.byte[2] &= 0xf7;  break;	
	case '_':	keyboard.byte[7] &= 0xfd;  keyboard.byte[4] &= 0xfe;  break;
	case ')':	keyboard.byte[7] &= 0xfd;  keyboard.byte[4] &= 0xfd;  break;
	case '(':	keyboard.byte[7] &= 0xfd;  keyboard.byte[4] &= 0xfb;  break;
	case '\'':	keyboard.byte[7] &= 0xfd;  keyboard.byte[4] &= 0xf7;  break;	
	case '&':	keyboard.byte[7] &= 0xfd;  keyboard.byte[4] &= 0xef;  break;
	case '%':	keyboard.byte[7] &= 0xfd;  keyboard.byte[3] &= 0xef;  break;
	case '$':	keyboard.byte[7] &= 0xfd;  keyboard.byte[3] &= 0xf7;  break;	
	case '#':	keyboard.byte[7] &= 0xfd;  keyboard.byte[3] &= 0xfb;  break;
	case '@':	keyboard.byte[7] &= 0xfd;  keyboard.byte[3] &= 0xfd;  break;
	case '!':	keyboard.byte[7] &= 0xfd;  keyboard.byte[3] &= 0xfe;  break;

// Sonderzeichen mittels CAPS SHIFT + Taste:
	case '':	keyboard.byte[0] &= 0xfe;  keyboard.byte[3] &= 0xfe;  break;		// EDIT
	case '':	keyboard.byte[0] &= 0xfe;  keyboard.byte[3] &= 0xfb;  break;		// TRUE VIDEO
	case '':	keyboard.byte[0] &= 0xfe;  keyboard.byte[3] &= 0xf7;  break;		// INV. VIDEO
	case 28:	keyboard.byte[0] &= 0xfe;  keyboard.byte[3] &= 0xef;  break;		// cursor left
	case 29: 	keyboard.byte[0] &= 0xfe;  keyboard.byte[4] &= 0xfb;  break;		// cursor right
	case 30:	keyboard.byte[0] &= 0xfe;  keyboard.byte[4] &= 0xf7;  break;		// cursor up
	case 31:	keyboard.byte[0] &= 0xfe;  keyboard.byte[4] &= 0xef;  break;		// cursor down
	case '}':	keyboard.byte[0] &= 0xfe;  keyboard.byte[4] &= 0xfd;  break;		// GRAPHICS
	case 8:	keyboard.byte[0] &= 0xfe;  keyboard.byte[4] &= 0xfe;  break;		// backspace

// Einzelzeichen mittels Taste:
	case 'v':	keyboard.byte[0] &= 0xef;  break;
	case 'c':	keyboard.byte[0] &= 0xf7;  break;	
	case 'x':	keyboard.byte[0] &= 0xfb;  break;
	case 'z':	keyboard.byte[0] &= 0xfd;  break;
	case 'g':   keyboard.byte[1] &= 0xef;  break;
	case 'f':	keyboard.byte[1] &= 0xf7;  break;	
	case 'd':	keyboard.byte[1] &= 0xfb;  break;
	case 's':	keyboard.byte[1] &= 0xfd;  break;
	case 'a':	keyboard.byte[1] &= 0xfe;  break;	
	case 't':   keyboard.byte[2] &= 0xef;  break;
	case 'r':	keyboard.byte[2] &= 0xf7;  break;	
	case 'e':	keyboard.byte[2] &= 0xfb;  break;
	case 'w':	keyboard.byte[2] &= 0xfd;  break;
	case 'q':	keyboard.byte[2] &= 0xfe;  break;
	case '5':   keyboard.byte[3] &= 0xef;  break;
	case '4':	keyboard.byte[3] &= 0xf7;  break;	
	case '3':	keyboard.byte[3] &= 0xfb;  break;
	case '2':	keyboard.byte[3] &= 0xfd;  break;
	case '1':	keyboard.byte[3] &= 0xfe;  break;
	case '6':   keyboard.byte[4] &= 0xef;  break;
	case '7':	keyboard.byte[4] &= 0xf7;  break;	
	case '8':	keyboard.byte[4] &= 0xfb;  break;
	case '9':	keyboard.byte[4] &= 0xfd;  break;
	case '0':	keyboard.byte[4] &= 0xfe;  break;
	case 'y':   keyboard.byte[5] &= 0xef;  break;
	case 'u':	keyboard.byte[5] &= 0xf7;  break;	
	case 'i':	keyboard.byte[5] &= 0xfb;  break;
	case 'o':	keyboard.byte[5] &= 0xfd;  break;
	case 'p':	keyboard.byte[5] &= 0xfe;  break;
	case 'h':   keyboard.byte[6] &= 0xef;  break;
	case 'j':	keyboard.byte[6] &= 0xf7;  break;	
	case 'k':	keyboard.byte[6] &= 0xfb;  break;
	case 'l':	keyboard.byte[6] &= 0xfd;  break;
	case 13:	keyboard.byte[6] &= 0xfe;  break;
	case 'b':   keyboard.byte[7] &= 0xef;  break;
	case 'n':	keyboard.byte[7] &= 0xf7;  break;	
	case 'm':	keyboard.byte[7] &= 0xfb;  break;
	case ' ':	keyboard.byte[7] &= 0xfe;  break;
	};

}


HandleKEYUP(Char)	 // global event enabling needed !!
{
	if (kbdMethod==iBasicKeys)  keyboard.quad.a = keyboard.quad.b = -1;
}


HandleAUTOKEY(Char) {}
