
/*	Handle preferences file

***	Copyright:	1995-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	09.Jan.95		Implemented load/save preferences			KIO !
	14.Jan.95		splitted this file from application.c			KIO !
	01.Apr.95 	major revision							KIO !
	15.Apr.95		separate preferences for m68 and ppc version	KIO !
*/

#include	<folders.h>
#include	<string.h>

#include	"kio.h"
#include	"strings.h"
#include	"preferences.h"
#include	"menue.h"
#include	"application.h"
#include	"messages.h"
#include	"z80.h"
#include	"movie.h"
#include	"cassette.h"
#include	"terminal.h"
#include	"display.h"
#include	"help+about.h"
#include	"z80.options"
#include	"sound.h"
#include	"z80 main.h"
#include	"keyboard.h"

#define	unlimited	1
#define	exact	2
#undef	highres			// from movie.h
#define	highres	3

#define	z80_t	1
#define	z80_ppc	2

#if GENERATINGPOWERPC
#define	resID			0x0101
#define	default_speed		exact
#define	default_engine		z80_ppc
#define	default_AY		on
#else
#define	resID			0x0100
#define	default_speed		exact
#define	default_engine		z80_t
#define	default_AY		off
#endif


#if !include_movie
Rect			movieRect		= { 0,0,100,200 };
WindowPtr 	movie_window=0;
#endif



// -----	Datenstruktur fr Preferences-Ressource
typedef struct 
{	Char		samplingRate;
	Char		Language;
	Char		Engine;
	Char		unused;
	Char		SoundVolume;
	Char		ScrnSize;
	Char		Speed;
	Char		KbdMethod;
	Rect		TapeRect;
	Rect		Specci1Rect;
	Rect		Specci2Rect;
	Rect		ConRect;
	Rect		KbdPictRect;
	Rect		Specci3Rect;
	Rect		Specci4Rect;
	Char		AYon;
	Char		force_offscreen;
	Rect		MovieRect;
	Char		Rest[54];				// padding to 128 bytes
} pfx;


short	pfxfile 		= 0;			// preferences file ID
short	oldresfile		= 0;			// active res file before opening pfxfile
pfx		prefs;					// preferences struct
FSSpec	pfxspec;					// file descriptor for open/create


// ----------------------------------------------------------------


// get portRect of window in global coordinates if window is open, otherwise use default
static Rect PortRect ( WindowPtr w, Rect r )
{	if (w) 
	{	SetPort ( w );
		r = w->portRect;
		LocalToGlobal ( (Point*)&r.top ); 
		LocalToGlobal ( (Point*)&r.bottom ); 
	};
	return r;
}	


//	Open or optionally create preferences file
//	after return, pfxfile contains the fileID (ok) or zero (error)
static _OpenPrefs (Boolean create)
{	OSErr err;

	if (sizeof(pfx)!=128)  { do_abort("\p sizeof(pfx)!=128 !!!","\p","\p"); return; }
	if (pfxfile) return;			// is already open (internal error!)
	oldresfile = CurResFile();		// remember actual resource file

//	find preferences folder & copy volume ref num and folder ref num -> pfxspec
	err = FindFolder ( kOnSystemDisk, 'pref', kCreateFolder, &pfxspec.vRefNum, &pfxspec.parID );
	if (err) return;				// preferences folder not found & unable to create

//	complete pfxspec by adding file name
	memcpy ( pfxspec.name, "\pMac Spectacle .prefs", 32 );

//	open preferences file
	pfxfile = FSpOpenResFile ( &pfxspec, fsRdWrPerm ); 
	if (pfxfile==-1)	 pfxfile=0;
	if (pfxfile || !create) return;

//	create preferences file
	FSpCreateResFile (&pfxspec, 'RSED', 'pref', 0 );	
	if (ResError()) return;
	PrintAlert(CStr(localized(" Preferences file was created"," Preferenzen-Datei wurde angelegt")));

//	and open it
	pfxfile = FSpOpenResFile ( &pfxspec, fsRdWrPerm ); 
	if (pfxfile==-1)	 pfxfile=0;
}


// close preferences file
static _ClosePrefs()
{	if (pfxfile) CloseResFile(pfxfile);
	pfxfile = 0;
	UseResFile(oldresfile);	
}



// read preferences from open file pfxfile -> struct prefs
static ReadPrefs()
{	Handle ppp;
	if (!pfxfile) return;		// prefs file not open
	UseResFile(pfxfile);

	ppp = GetResource ('pref', resID);
	if (ppp) 
	{	HLock(ppp);
		memcpy ( &prefs,*ppp,sizeof(pfx));
		HUnlock(ppp);
		ReleaseResource(ppp);
	};
}


// write preferences to open pfxfile
static WritePrefs()
{	Handle ppp;
	if (!pfxfile) return;		// prefs file not open
	UseResFile(pfxfile);

	ppp = GetResource ('pref', resID);
	if (!ppp)				// resource does not exist => create
	{	ppp = NewHandle ( sizeof(pfx) );
		if (ppp)
		{	AddResource ( ppp, 'pref', resID, "\p" );
			if (ResError()) { DisposHandle(ppp); ppp=nil; }
		}
	}

	if (ppp)
	{	HLock(ppp);
		memcpy ( *ppp,&prefs,sizeof(pfx));
		HUnlock(ppp);
		ChangedResource (ppp);
		ReleaseResource(ppp);
	};
}


// Initialize prefs struct with default values
static InitPrefs()
{	memset ( &prefs, 0, sizeof(pfx) );
//	prefs.samplingRate	= 0;
//	prefs.SoundVolume	= 0;
	prefs.AYon		= default_AY;
	prefs.force_offscreen=0;
	prefs.Language		= iInternational;
	prefs.unused		= 0;
	prefs.ScrnSize		= 2;
	prefs.KbdMethod	= iGameKeys;
	prefs.Speed		= default_speed;
	prefs.Engine		= default_engine;
//	SetRect (&theRect, rLeft, rTop, rRight, rBottom);
	SetRect (&prefs.MovieRect, 	7, 150, 250+7, 102+150);
	SetRect (&prefs.TapeRect, 	7, 45,   250+7, 102+45);
	SetRect (&prefs.ConRect,		7, 50,   460+7, 360+50);

	SetRect (&prefs.Specci1Rect, 	10, 40, 10+1*(256+48), 40+1*(196+32));
	SetRect (&prefs.Specci2Rect, 	10, 40, 10+2*(256+48), 40+2*(196+32));
	SetRect (&prefs.Specci3Rect, 	10, 40, 10+3*(256+48), 40+3*(196+32));
	SetRect (&prefs.Specci4Rect, 	10, 40, 10+4*(256+48), 40+4*(196+32));

	SetRect (&prefs.KbdPictRect, 	8, 300, 632, 366);
}


// Check all values in struct prefs
static CheckPrefs()
{	prefs.samplingRate &= 0x01;
	if (prefs.Language!=iInternational&&prefs.Language!=iGerman) prefs.Language=iInternational;
	if (prefs.unused!=0) prefs.unused=0;
	if (prefs.ScrnSize<1 || prefs.ScrnSize>4) prefs.ScrnSize=4;
	if (prefs.Speed<1||prefs.Speed>3) prefs.Speed=default_speed;
	if (prefs.Engine<1||prefs.Engine>2) prefs.Engine=default_engine;
	if (prefs.KbdMethod!=iBasicKeys&&prefs.KbdMethod!=iGameKeys) prefs.KbdMethod=iBasicKeys;
	//if (prefs.SoundVolume<0||prefs.SoundVolume>7) prefs.SoundVolume=0;
// rectangles should be checked too ...
	if (prefs.MovieRect.right<=prefs.MovieRect.left) 	SetRect (&prefs.MovieRect, 	7, 150, 250+7, 102+150);
}


// 	Initialize all settings which are stored in the preferences file
// 	call this procedure after setting up the menu bar
// -----	Open resource fork of preferences file
//		If the file doesn't exist, it is created
//		note: all resources in this file obscure identical resources in the application
ReadPreferences()
{
	InitPrefs();
	_OpenPrefs(false);
	ReadPrefs();
	_ClosePrefs();
	CheckPrefs();

	tapeRect		= prefs.TapeRect;
	movieRect		= prefs.MovieRect;
	terminalrect	= prefs.ConRect;
	specciRect[1]	= prefs.Specci1Rect;
	specciRect[2]	= prefs.Specci2Rect;
	specciRect[3]	= prefs.Specci3Rect;
	specciRect[4]	= prefs.Specci4Rect;
	kbdPictRect	= prefs.KbdPictRect;

	TermOpen();	// open con window before opening specci window!

	HandleMENU ( mOptions,	prefs.Language );
	HandleMENU ( mOptions,	iLogOnlyCon );
	HandleMENU ( mOptions,	iQuiet );
	HandleMENU ( mLogging,	iLogNoLegals );
	HandleMENU ( mOptions,	prefs.KbdMethod );
	HandleMENU ( mOptions,	iNormalSize-1+prefs.ScrnSize );
	HandleMENU ( mOptions,	prefs.force_offscreen ? iForceOffscreen : iAvoidOffscreen );

	SetSampleRate ( prefs.samplingRate ? sample_rate_44 : sample_rate_22 );
	SetSoundVolume ( prefs.SoundVolume );
	SetSoundChannels ( prefs.AYon );

	HandleMENU ( mSpeed,	iUnlimitedSpeed-1+prefs.Speed );
	HandleMENU ( mSpeed,	iBetter-1+prefs.Engine );
}



WritePreferences()
{	
	prefs.TapeRect		= PortRect ( tape_window, tapeRect );
	prefs.MovieRect	= PortRect ( movie_window, movieRect );
	prefs.ConRect		= PortRect ( terminalwindow, terminalrect );
	prefs.KbdPictRect	= PortRect ( helpscreen, kbdPictRect );
	prefs.Specci1Rect	= PortRect ( scrnsize==iNormalSize?zxwindow:nil, specciRect[1] );
	prefs.Specci2Rect	= PortRect ( scrnsize==iDoubleSize?zxwindow:nil, specciRect[2] );
	prefs.Specci3Rect	= PortRect ( scrnsize==iTripleSize?zxwindow:nil, specciRect[3] );
	prefs.Specci4Rect	= PortRect ( scrnsize==iQuadSize?zxwindow:nil, specciRect[4] );

	prefs.SoundVolume	= GetSoundVolume();
	prefs.samplingRate	= GetSampleRate()==sample_rate_44;			// 22kHz=0 : 44 kHz=1
	prefs.AYon		= GetSoundChannels() != 0;
	prefs.force_offscreen=force_offscreen;
	prefs.unused		= 0;
	prefs.ScrnSize		= scrnsize;
	prefs.Language		= language;
	prefs.KbdMethod	= kbdMethod;
	prefs.Speed		= run_unlimited+2*run_exact+3*run_highres;
	prefs.Engine		= use_better_z80+2*use_best_z80;

	_OpenPrefs(true);
	WritePrefs();
	_ClosePrefs();
}




