

/*	Logfile, messages and alerts

***	Copyright:	1994-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	31.Mar.96:	Extracted this file from Mac Spectacle's application.c		KIO !
*/


#include	"kio.h"
#include	"strings.h"
#include	"terminal.h"
#include	"messages.h"


Boolean	log_file	= false;			// loging to file enabled ?
Boolean	log_con	= true;			// loging to console enabled ?
short	log_level	= msg;			// verbosity of log messages

short	log_ID	= 0;				// log file reference number
long		log_crea	= 'R*ch';			// creator for log file


// ***** LOGGING, MESSAGES AND ALERTS ***************************************** 



// -----	Print  Character  to Logfile
#define	LogChar(Z)									\
{	static long N = 1;  char C = Z;							\
	if (log_ID)  if ( FSWrite ( log_ID, &N, &C ) )  AbortLogfile();	\
}


// -----	close the logfile after an error
AbortLogfile();
AbortLogfile()
{
	log_file = false;
	log_con = true;

	if (log_ID)
	{	FSClose(log_ID); log_ID=0;
		PrintXAlert ( CStr("Error while writing to logfile"), CStr("The logfile has been closed") );
	};

	UpdateLogMenu();
}


// -----	close the logfile
void CloseLogfile()
{
	log_file = false;
	log_con = true;
	
	if (log_ID)
	{
		FSClose(log_ID); log_ID=0;
		PrintLine(CStr("----- Logfile closed -----"));
	}

	UpdateLogMenu();
}


// -----	enable logging to file
void OpenLogfile()
{	
	if (log_file) return;

	if ( !log_ID )
	{	static StandardFileReply reply;
		static Boolean reply_valid = false;
		OSErr err;
	
		if (reply_valid)
		{	
			if (do_confirm("\pappend to old logfile?",0)==1)
			{
				reply_valid=false;
				err = FSpOpenDF ( &reply.sfFile, fsCurPerm, &log_ID );
				if (err) { PrintAlert ( CStr("Error opening logfile") ); goto x; };
				err=SetFPos ( log_ID, fsFromLEOF, 0 );	
				if (err) { PrintAlert ( CStr("Error positioning logfile") ); goto x; };
				reply_valid=true;
				log_file = true;
				UpdateLogMenu();
				LogChar(13);
				PrintLine(CStr("+++++ Logfile open +++++"));
				LogChar(13);
				return;
			}
		}
		
	 	reply_valid=false;
	 	StandardPutFile ( "\pLogfile Name:", "\plogfile.txt", &reply );
		if (!reply.sfGood) goto x;
		err = FSpCreate ( &reply.sfFile, log_crea, 'TEXT', 0 ); if (err==dupFNErr) err=noErr;
		if (err) { PrintAlert ( CStr("Error creating logfile") ); goto x; };
		err = FSpOpenDF ( &reply.sfFile, fsCurPerm, &log_ID );
		if (err) { PrintAlert ( CStr("Error opening logfile") ); goto x; };
		err=SetFPos ( log_ID, fsFromStart, 0 );	
		if (err) { PrintAlert ( CStr("Error positioning logfile") ); goto x; };
		reply_valid=true;
	};
		
	PrintLine(CStr("+++++ Logfile open +++++"));
	log_file = true;
	UpdateLogMenu();
	return;

x:	log_file = false;
	log_con = true;
	log_ID = 0;
	UpdateLogMenu();
}


// -----	Print NL  to Console and/or Logfile
PrintNL ()
{	if (log_con)	TermNL();
	if (log_file)	LogChar(13);
}


// -----	Print Character to Console and/or Logfile
PrintChr ( Char c ) 
{	if (log_con)	TermChar(c);
	if (log_file)	LogChar(c);
}


// -----	Print Character + Space to Console and/or Logfile
PrintChar ( Char c )
{	if (log_con)	{ TermChar(c); TermSpace(); }
	if (log_file)	{ LogChar(c); LogChar(32); }
}


// -----	Print String + Space to Console and/or Logfile
PrintStr ( string s )	
{	
	if (log_con) { TermWrap(s); TermSpace(); }	
	if (log_file) 
	{	if (log_ID)  if ( FSWrite ( log_ID, &s.len, s.text ) )  AbortLogfile(); 	
		LogChar(' ');
	}
}


// -----	Print 2/4/8 Digit Hex Number to Logfile and/or Console
PrintHex2 (long n)	{	PrintStr ( HexStr(n,2));	}
PrintHex4 (long n)	{	PrintStr ( HexStr(n,4));	}
PrintHex8 (long n)	{	PrintStr ( HexStr(n,8));	}


// -----	Print Line ----------------------------------------------------
PrintLine ( string s )	
{	PrintStr ( s );
	PrintNL();
}

PrintMsg ( string s )
{	if (log_level<=msg) PrintLine(s);
}

PrintDebug ( string s )
{	if (log_level<=debug) PrintLine(s);
}

PrintAlert( string s )
{	
	do_alert(PascalStr(s),0,0);
	PrintLine(s);
}

PrintNumberAlert( string s, long n )
{	
	do_alert(PascalStr(s),PascalStr(NumStr(n)),0);
	PrintStr(s);
	PrintLine(NumStr(n));
}

PrintXAlert( string s1, string s2 )
{	
	do_alert(PascalStr(s1),PascalStr(s2),0);
	PrintLine(s1);
	PrintLine(s2);
}



