
/*	Z80 Emulator

***	Copyright:	1994-2000 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	This file must be #included in Z80.m68.c to work
	It contains the body of Z80()
*/



// -----	Register usage -------------------------------------------------------------
#define	target	a0		// general use as target pointer

#define	ai		a1		// index register pointer in ix/iy cmnd processing
#define	ir		(a1)		// access ix or iy
#define	ih		(a1)		// access XH or YH
#define	il		1(a1)	// access XL or YL
#define	air		-2(a1)	// access AIX or AIY

#define	ip		a2		// instruction pointer:	 zreg.IP.ptr
#define	rp		a3		// return stack pointer:	 zreg.RP.ptr
#define	loop		a4		// -> main instruction dispatcher
#define	logflags	a6		// -> array mlog_flags[]

#define	scratch	d0
#define	scratch2	d1

#define	rom_end	d2
#define	core		d3		// upper half = CORE base address
#define	db		d4		// 0x000000XX:	upper 3 bytes are always 0; db MUST be D4 
#define	a		d5		// 0x00AA00AA:	a.b = A; upper half = A'
#define	f		d6		// 0x00FF00FF:	f.b = F; upper half = F'
#define	cc		d7		// T cycle counter:	count down to 0

#define	m_next	jmp	(loop)


// -----	exact timing ------------------------------------------------------
#define	used_regs		a2-a4/a6/d3-d7
#define	sto_cc		move.l	cc,CYCLES
#define	ld_cc		move.l	CYCLES,cc
#define	more1		subq.l	#1,cc	
#define	more2		subq.l	#2,cc	
#define	more3		subq.l	#3,cc	
#define	more4		subq.l	#4,cc	
#define	more5		subq.l	#5,cc	
#define	more6		subq.l	#6,cc	
#define	more7		subq.l	#7,cc	
#define	more8		subq.l	#8,cc	


// -----	option depending code macros ----------------------------------------
#if rom_protection
#define	load_rom_end	move.l  core,rom_end;    move.w  #0x4000,rom_end
#define	if_rom(LBL)	cmpa.l  rom_end,a0;    bcs.s  LBL
#else
#define	load_rom_end
#define	if_rom(LBL)
#endif


// =====	Z80 ENGINE BODY ===========================================

//short asm Z80 ( )	{
	

// save registers
	movem.l	used_regs,-(a7)

// load frequently needed constants into mc68020 registers:
	move.l	CORE,core	
	moveq	#0,db
	lea		mlog_flags,logflags
	lea		nxtcmnd,loop

// load lreg struct
	lea		zreg,a0
	lea		lreg,a1
	move.w	(a0)+,core;	move.l	core,(a1)+	// bc
	move.w	(a0)+,core;	move.l	core,(a1)+	// de
	move.w	(a0)+,core;	move.l	core,(a1)+	// hl
	move.w	(a0)+,core;	move.l	core,(a1)+	// ix
	move.w	(a0)+,core;	move.l	core,(a1)+	// iy

// load the Z80 registers, which are held in mc68020 registers:
	ld_cc					// T cycles
	load_r					// if exact_r
	load_rom_end				// if rom_protection

// load A and A'
	moveq.l	#0,a
	move.b	RA2,a			// A2
	swap 	a
	move.b	RA,a				// A

// load F and F'
	lea		m68flags,a0
	moveq.l	#0,f
	move.b	RF2,f			// F2
	move.b	(a0,f.w),f			// konvert F2 to m68 layout
	swap 	f
	move.b	RF,f				// F
	move.b	(a0,f.w),f			// konvert F to m68 layout

// load PC
	move.w	PC,core
	move.l	core,ip			// PC

// load SP
	move.w	SP,core
	movea.l	core,rp
	bne.s 	spok
	adda.l	#0x00010000,rp;	// if sp==$0000
spok:


// check WUFF to see, whether an NMI or IRPT is pending:
check_wuff:
	tst.b 	WUFF
	beq		nxtcmnd			// no flag set
	bmi.s 	handle_nmi		// bit7: nmi
//	bgt.s 	handle_irpt		// bit0-6: irpt


// ------	Interrupt dispatcher --------------------------------------------------
handle_irpt:
	tst.b		IFF1
	beq		nxtcmnd			// interrupts are disabled
	do_info_irpt
	subq.b	#1,WUFF			// clear interrupt request
	clr.w	IFF				// disable iff1 & iff2
	cmp.b	#1,IM			// which irpt mode ?
	bgt.s		im2				// irpt mode 2:  indirect call
	blt.s  	im0				// irpt mode 0:  read irpt instruction
im1:	move.w	#56,core		// ---- irpt mode 1:  rst 56
j13:	more5					// 13 T cycles
j08:	more8					// 
j00:	move.w	ip,d0				// push PC
	rol.w	#8,d0			// ...
	move.w	d0,-(rp)			// ...
	movea.l	core,ip			// load irpt routine address
	m_next					// next instruction

im2:	more6				// ---- irpt mode2:  19 T cycles ((6+8+5))
	move.w	RI,core			// read both: i-register & IRPTCMD !!!
	movea.l	core,a0
	move.w	(a0),core
	rol.w	#8,core
	bra.s 	j13

im0:	more5				// ---- irpt mode 0:  1 T cycle + normal instruction timing
	move.b	IRPTCMD,db		// irpt mode 0 => read instruction from bus
	bra		xx_im0			// perform instruction (multi-byte instructions not supported!)
	
handle_nmi:				// ---- nmi
	do_info_nmi
	more3					// nmi: 11 T cycles  ((3+8))
	move.b	IFF1,IFF2			// save iff1
	sf		IFF1				// di
	bclr		#7,WUFF			// clear nmi flag
	move.w	#0x0066,core		// nmi address
	bra.s		j08


// -----	T cycle counter reached 0 --------------------------------------
cycle00:	
	addq.l	#4,cc			// undo last subq
	sto_cc					// update CYCLES
	jsr		Do_Cycles 		// do whatever must be done
	load_rom_end
	ld_cc					// get CYCLES again
	tst.b		EXIT
	bne		exit_watchdog		// watchdog => exit!
	tst.b 	WUFF
	beq.s 	nxtcmnd			// no flag set
	bgt	 	handle_irpt		// irpt
	blt.s		handle_nmi		// nmi


// =====	BRANCHING =========================================


xx32:	btst	#2,f;	beq.s 	xx24;	more3;	addq.l	#1,ip;	m_next	// JR NZ,dis
xx40:	btst	#2,f;	bne.s 	xx24;	more3;	addq.l	#1,ip;	m_next	// JR Z,dis
xx48:	btst	#0,f;	beq.s 	xx24;	more3;	addq.l	#1,ip;	m_next	// JR NC,dis
xx56:	btst	#0,f;	bne.s 	xx24;	more3;	addq.l	#1,ip;	m_next	// JR C,dis

xx16a:	addq.l #1,ip;	more4;	m_next
xx16:	subq.b #1,RB;	beq.s xx16a;	more1;	// ---- DJNZ dis: Handler (( the most frequent instruction! ))
		
xx24:					// JR dis
	more8				// 12 T cycles if branched; except DJNZ: 13 T cycles
	move.b	(ip)+,core
	ext.w	core
	add.w	ip,core

nxtjmp:					// ---- complete jump instruction
	movea.l	core,ip


// -----	THE MAIN INSTRUCTION DISPATCHER --------------------------------

nxtcmnd:
#include	"cmd_xx.c"		// used for all instructions which do not modify the PC 
#include	"cmd_CB.c"		// cmd_CB:
#include	"cmd_ED.c"		// cmd_ED:
#include	"cmd_IX/iy.c"		// cmd_IX:  and  cmd_IY:
#include	"cmd_IX/IY_CB.c"	// XYCB:
#include	"cmd_DAA.c"


// -----	Exit points -------------------------------------------------------------

exit_weird:	moveq	#weird_instr,d0		// informational exit
			bra.s 	exit_z80

exit_ill:		moveq	#ill_instr,d0			// informational exit
			bra.s 	exit_z80

exit_nimp_2:	subq.l	#1,ip				// abort:	otir/otdr/inir/indr
exit_nimp_1:	subq.l	#1,ip				// abort:	daa
			moveq	#nimp_instr,d0
			bra.s	 	exit_z80

exit_rst0:	moveq	#rst0_instr,d0			// patch: rst 0
			bra.s 	exit_z80

exit_halt:		moveq	#halt_instr,d0			// halt
			bra.s 	exit_z80

exit_watchdog:	sf		EXIT					// clear watchdog flag
			moveq	#watchdog_irpt,d0		// select return code

exit_z80:		move.w	ip,PC
			move.w	rp,SP

			move.b	a,RA					// A
			swap		a
			move.b	a,RA2				// A2

			lea		z80flags,a0
			move.b	(a0,f.w),f				// konvert F to z80 layout
			move.b	f,RF					// F
			swap 	f
			move.b	(a0,f.w),f				// konvert F2 to z80 layout
			move.b	f,RF2				// F2

			sto_cc						// CYCLES
			store_r						// r (optional)
			movem.l	(a7)+,used_regs		// restore old m68 registers

			lea		zreg,a0				// write back local registers
			lea		lreg,a1
			move.w	2(a1),(a0)+
			move.w	6(a1),(a0)+
			move.w	10(a1),(a0)+
			move.w	14(a1),(a0)+
			move.w	18(a1),(a0)+

			rts

//	}	end of Z80()


#undef	rom_protection
#undef	more1
#undef	more2
#undef	more3
#undef	more4
#undef	more5
#undef	more6
#undef	more7
#undef	more8
#undef	sto_cc
#undef	ld_cc
#undef	if_rom
#undef	used_regs
#undef	load_rom_end
#undef	exact_bit
#undef	exact_xy_bit




