/*	KIO's standard header file.

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
*/

#ifndef	KIO_h
#define	KIO_h KIO_h

#ifndef COMPILINGPOWERPC
#define COMPILINGPOWERPC true
#endif


// ---- Basic data type definitions ------------------------------------------
typedef	unsigned char		Char;
typedef	unsigned short		Short;
typedef	unsigned long		Long;
typedef	long long			llong;
typedef	unsigned long long	LLong;
#define	bool				unsigned char


// -----	Constants -----------------------------------------------------------
#define 	true			1
#define 	false 			0
#define 	yes			1
#define 	no			0
#define 	on			1
#define 	off			0


// -----	Error codes ---------------------------------------------------------
enum
{		error		= -1,
		noerror		=  0,
		internalerror,
		outofmemory,
		notanumber,
		numbertoobig,
		notimplemented,
		parametererror,
		last_error_id		// custom errors may be defined starting here e.g. in "options.h"
};


// ----	Resource id's -------------------------------------------------------
const int		abort_id		= 666;				// DITL and ALRT for bomb alert
const int		stop_id		= 667;				// DITL and ALRT for stop alert
const int		warning_id	= 668;				// DITL and ALRT for warn box
const int		message_id		= 669;				// DITL and ALRT for info box
const int		yesno_id		= 670;				// DITL and ALRT for yes/no requester
const int		yesnocancel_id	= 671;				// DITL and ALRT for yes/no/cancel requester
const int		resumeabort_id	= 672;				// DITL and ALRT for resume/abort requester


// ----	Global variables ----------------------------------------------------
#define		abort			MYABORT				// naming problem
extern		OSErr		err;					// central error dump
extern		EventRecord	event;				// event returned by PollEvents()
extern		bool	 		done;					// quit application ?	( 'apple'+'Q' )
extern		bool			abort;				// abort operation ?	( 'apple'+'.' )
extern		bool			powermgr;				// Power manager present ?
extern		bool			initappl;				// InitAppl() called.


// -----	Nice defines -------------------------------------------------------
#define		MIN(a,b)		((a)<(b)?(a):(b))
#define		MAX(a,b)		((a)>(b)?(a):(b))
#define		SWAP(a,b)	((a)^=(b),(b)^=(a),(a)^=(b))	//  (c) KIO !
#define		SIGN(a)		(((a)>0)-((a)<0))			//  (c) KIO !
#define		SORT(a,b)		if ((a)<=(b)) {} else SWAP(a,b)

inline long		Min			( long a, long b )			{ return a<b ? a : b; }
inline long		Max			( long a, long b )			{ return a>b ? a : b; }	
inline long		MinMax		( long a, long n, long e )	{ return n<=e ? (n>=a?n:a) : e; }

inline int		Sign			( long a )				{ return SIGN(a); }
inline long		Abs			( long a )				{ return a>0 ? a : -a; }

inline void		Swap			( long& a, long& b )		{ long c=a; a=b; b=c; }
inline void		Swap			( void*& a, void*& b )	{ void*c=a; a=b; b=c; }
inline void		Sort			( long& a, long& b )		{ if (a>b) SWAP(a,b); }

inline long		Random		( Long sz )				{ return ((Long)(Short)Random())*sz/0x0ffff; }

inline long		Scale			( long oldval, long oldmax, long newmax )	// size conversion ((caveat: oldmax*newmax <= 0x7FFFFFFF))
											{ return ( oldval*newmax+oldmax/2 ) / oldmax; }

inline void 		EnableItem 	( MenuHandle m, int i, bool f )
										 	{ if (f) EnableItem (m,i);  else DisableItem(m,i); }

inline bool		InBox		( Rect b, Point p )		{ return ( p.h>=b.left && p.h<b.right && p.v>=b.top && p.v<b.bottom ); }


// -----	Read options --------------------------------------------------------
#include	"options.h"

#define	XSAFE	(SAFE>=1)
#define	XXSAFE	(SAFE>=2)
#define	XLOG		(LOG>=1)
#define	XXLOG	(LOG>=2)

#ifndef	KIOS_THREADS
#error	please define KIOS_THREADS true or false!
#endif

#if		KIOS_THREADS
#include	"dispatcher.h"
#define 	ERR 			Me()->err
#define	DISP()		Disp()
#else
#define	ERR			err
#define	DISP()		
#endif


// -----	malloc.cp ----------------------------------------------------------
#include	<mcompile.h>
//void* 		operator new 	( size_t size );		// it just makes no difference 
//void		operator delete	( void* p );		//  whether defined here or not
inline char*	MAlloc		( long size )		{ return new char[size]; }	// for backward compatibility
inline void		MFree		( void* ptr )		{ delete [] (char*)ptr; }

#if MALLOC_STATISTICS
extern long		malloc_xrequests;	// total number of xmem requests (internal and external)
extern long		malloc_xmemsize;	// memory currently allocated externally
extern long		malloc_xmemblocks;	// number of blocks currently allocated externally

extern long		malloc_requests;	// total number of malloc requests
extern long		malloc_size;		// memory currently in use
extern long		malloc_blocks;		// number of blocks currently in use
#endif


// -----	misc. files -------------------------------------------------------
void			InitAppleEvents	( );
void			CheckGestalt	( );
Char* 		ErrorText 		( OSErr );
void			InitAppl		( );


// -----	mem.cp ------------------------------------------------------------
#include	"mem.h"


// -----	quickstrings.cp -------------------------------------------------------
#include	"quickstrings.h"


// -----	logfile.cp -----------------------------------------------------------
#include	"logfile.h"


// -----	alerts.cp -------------------------------------------------------------
short			DoAlert		( short n, const Char*s, const Char*t=0, const Char*u=0, const Char*v=0 );
void			DoAbort		( const Char*s,const Char*t=0,const Char*u=0,const Char*v=0 );
bool			ConfirmResume	( const Char*s,const Char*t=0,const Char*u=0,const Char*v=0 );	// -> resume or abort
bool			Confirm		( const Char*s,const Char*t=0,const Char*u=0,const Char*v=0 );	// -> yes or no
 
void	 		_trap_error_	( char*file, long line, OSErr );
extern OSErr	_err_;
#define		TRAP(XXX)	if (!(_err_=(XXX))) {} else _trap_error_(__FILE__,__LINE__,_err_)
#define		XTRAP(XXX)	if (!XSAFE) 	    {}	else TRAP(XXX)
#define		XXTRAP(XXX)	if (!XXSAFE) 	    {}	else TRAP(XXX)
#define		OMEM(XXX)	if (XXX) 		    {}	else	_trap_error_(__FILE__,__LINE__,outofmemory)
#define		IERR()		_trap_error_(__FILE__,__LINE__,internalerror)


// -----	windows.cp -------------------------------------------------------------
void			DisplayWindow	( WindowPtr w );
void			SetAlertRect 	( Rect *box, short w, short h );


// -----	mouse pointer.cp -------------------------------------------------------
void			AnimateCursor	( int res_id=0, int frames=0 );
void			WaitCursor   	( );
void			ArrowCursor  	( );


// -----	pollevents.c -------------------------------------------------------
#define		hsb_magic		'HsbM'
#define		vsb_magic		'VsbM'
Boolean		PollEvents 		( short evMask, long evDelay, RgnHandle evRegion );
void			ScrollWindowRect ( WindowPtr window, Rect *box, short dx, short dy );
void			ResizeWindow	( WindowPtr window, short x0, short y0, short w, short h );
ControlHandle	FindRefControl	( WindowPtr window, long refValue );


// -----	preferences.c ------------------------------------------------------
struct		String;
OSErr		ClosePrefs		();
OSErr		OpenPrefs		();
OSErr		GetPref		(long id, String& s);
OSErr		PutPref		(long id, const String& s);
OSErr		GetSecurePref	(long id, String& s);
OSErr		PutSecurePref	(long id, const String& s);
OSErr		PutHexPref	(long id, const Char *p, long n);
OSErr		GetHexPref		(long id, Char *p, long n);


// these procedures must be provided somewhere in your application:
void		HandleOPEN		(FSSpec);
void		HandlePRINT		(FSSpec);
void		HandleNEW			();
void		HandleDISK		();
void		HandleDRIVER		();
void		HandleAPP3		();
void		HandleKEYDOWN		(Char);
void		HandleAUTOKEY		(Char);
void		HandleKEYUP		(Char);
void		HandleMENU		(short, short);
void		HandleCLOSE		(WindowPtr);
void		HandleGROW		(WindowPtr);
void		HandleZOOM		(WindowPtr);
void		HandleDRAG		(WindowPtr,Point);				// only for Mac Spectacle ...
void		HandleINCONTENT	(WindowPtr);
void		HandleBUTTON		(WindowPtr,ControlHandle);
void		HandleCHECKBOX		(WindowPtr,ControlHandle);
void		HandleSLIDEBAR		(WindowPtr,ControlHandle,short);
void		HandleMOUSEUP		();
void		HandleACTIVATE		(WindowPtr);
void		HandleDEACTIVATE	(WindowPtr);
void		HandleUPDATE		(WindowPtr);
void		HandleFASTUPDATE	(WindowPtr);
void		HandleRESUME		();
void		HandleSUSPEND		();
void		HandleREGION		();
void		HandleNULL		();



#endif


