
/*	application dependent z80 debugger call handlers

***	Copyright:	1995-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	23.01.95:	Created this file							KIO !
    20.Oct.01   revised for CW7Pro and ANSI C compliance	hdusel@bnv-gz.de
*/

#include	<events.h>

#include	"kio.h"
#include	"strings.h"
#include	"z80.h"
#include	"z80 main.h"
#include	"disassembler.h"
#include	"application.h"
#include	"messages.h"
#include	"z80_opcodes.h"
#include	"debugger.h"
#include	"menue.h"

Boolean		show_1st_instr	= false;	// Z80_1st_Instr(cc,ip)	log first use of every instruction?
Boolean		show_1st_location	= false;	// Z80_1st_Loc(cc,ip)	log first execution of every location?

Boolean		show_illegals		= false;	// Z80_Illegals(cc,ip)	log use of illegal instructions?
Boolean		stop_on_illegals	= false;	//					halt engine?

Boolean		show_weird		= false;	// Z80_Weird(cc,ip)		log use of weird illegal instructions?
Boolean		stop_on_weird		= false;	//					halt engine?

Boolean		show_irpt			= false;	// Z80_Info_Irpt(cc)	log Booleanerrupt execution?
Boolean		show_nmi			= false;	// Z80_Info_NMI(cc)	log nmi execution?

Boolean		show_ei			    = false;	// Z80_Info(cc,ip)		log use of special instructions?
Boolean		show_di			    = false;
Boolean		show_im0			= false;
Boolean		show_im1			= false;
Boolean		show_im2			= false;
Boolean		show_reti			= false;
Boolean		show_retn		= false;
Boolean		show_halt			= false;
Boolean		show_daa			= false;
Boolean		show_r			= false;
Boolean		show_i			= false;
Boolean		show_rst0		= false;
Boolean		show_rst1		= false;
Boolean		show_rst2		= false;
Boolean		show_rst3		= false;
Boolean		show_rst4		= false;
Boolean		show_rst5		= false;
Boolean		show_rst6		= false;
Boolean		show_rst7		= false;

Boolean		enable_ROM_write	= false;
Boolean		stop_on_ROM_write	= false;


// ------------------------------------------------------------------------------

void	PrintInfo ( Short ip );
void	PrintInfo ( Short ip )
{
//	PrintHex8	( TOTAL-CYCLES );			// time stamp
	PrintHex4	( ip );					// PC
	PrintLine ( Disassemble(&ip) );			// caveat: increments ip !!
}


// -----	information call for illegal instructions -------------------------------------
//		called for SLL and standard illegal opcodes using XH, XL, YH and YL
//		returns flag which requests Z80() to exit
Boolean Z80_Info_Illegal ( long cc, Short ip )
{	CYCLES = cc;
	if ( show_illegals )		PrintInfo ( ip );
	return stop_on_illegals;
}
	

// -----	information call for illegal instructions -------------------------------------
//		called for illegal instructions except SLL and standard illegal opcodes using XH, XL, YH and YL
//		returns flag which requests Z80() to exit
Boolean Z80_Info_Weird ( long cc, Short ip )
{
	if ( Peek(ip)==0xED )
	{	switch ( Peek(ip+1) )
		{
// xzx:	ED 00 Dumps the current processor state, and wait for a keypress
//		to continue.  If 'q' is pressed, xzx will exit.
		case 0x00:	break;

// xzx:	ED FB activates the level loader trap.
//		loads  "actualgameX.DAT"  at location stored in register HL
//		actualgame = actual snapshot name (without extension)
//		X = required level  =  register A
		case 0xFB:	break;

// xzx:	ED FC - Activates the tape loader trap.
//		this is placed at 0x0556 and loads a header plus data block from a single file
//		i don't think i'll support that.
		case 0xFC:	break;

// xzx:	ED FD - Activates the tape saver trap.
//		this is placed at 0x04C2  (+3: 0x04C6) and stores one header plus it's data block in a single file
//		i don't think i'll support that.
		case 0xFD:	break;

// xzx:	ED FE - Activates RS232 input.  A byte read from stdin is placed in register A.
		case 0xFE:	break;

// xzx:	ED FF - Activates RS232 output.  Register A is written to stdout.
		case 0xFF:	break;
		}
	}

	CYCLES = cc;
 	if ( show_weird )	PrintInfo ( ip );
	return stop_on_weird;
}
	

// -----	miscellaneous instructions info call ----------------------------------------

void Z80_Info ( long cc, Short ip )
{	
	CYCLES = cc;

	switch ( Peek(ip) )
	{
	case	EI:		if (show_ei)	PrintInfo ( ip );	return;
	case DI:		if (show_di)	PrintInfo ( ip );	return;
	case DAA:		if (show_daa)	PrintInfo ( ip );	return;
	case HALT:	if (show_halt)	PrintInfo ( ip );	return;
	case RST00:	if (show_rst0)	PrintInfo ( ip );	return;
	case RST08:	if (show_rst1)	PrintInfo ( ip );	return;
	case RST10:	if (show_rst2)	PrintInfo ( ip );	return;
	case RST18:	if (show_rst3)	PrintInfo ( ip );	return;
	case RST20:	if (show_rst4)	PrintInfo ( ip );	return;
	case RST28:	if (show_rst5)	PrintInfo ( ip );	return;
	case RST30:	if (show_rst6)	PrintInfo ( ip );	return;
	case RST38:	if (show_rst7)	PrintInfo ( ip );	return;
	case 0xED:
		switch ( Peek(ip+1) )
		{	case ED4E:
			case ED66:
			case ED6E:
			case IM_0:	if (show_im0)	PrintInfo ( ip );	return;
			case ED76:
			case IM_1:	if (show_im1)	PrintInfo ( ip );	return;
			case ED7E:
			case IM_2:	if (show_im2)	PrintInfo ( ip );	return;
			case ED5D:
			case ED6D:
			case ED7D:
			case RETI: 	if (show_reti)	PrintInfo ( ip );	return;
			case ED55:
			case ED65:
			case ED75:
			case RETN: 	if (show_retn)	PrintInfo ( ip );	return;
			case LD_A_I:
			case LD_I_A:	if (show_i)	PrintInfo ( ip );	return;
			case LD_A_R:
			case LD_R_A:	if (show_r)	PrintInfo ( ip );	return;
		}
	}

	PrintInfo ( ip );		// this should never happen !
}


// -----	Log every interrupt processing ---------------------------------
//		called before processing interrupt
//		informational call only!
void Z80_Info_Irpt( long cc )
{	if ( !show_irpt ) return;
	PrintHex8 ( TOTAL-cc );	// time stamp
	PrintLine	(PStr("\pInterrupt"));
}


// -----	Log every nmi processing -------------------------------------
//		called before processing nmi
//		informational call only!
void Z80_Info_NMI ( long cc )
{	if ( !show_nmi ) return;
	PrintHex8 ( TOTAL-cc );	// time stamp
	PrintLine	(PStr("\pNMI"));
}


// -----	Log first execution of any location -----------------------------
//		called before executing instruction and after decrementing cc by 4
//		informational call only!
void Z80_1st_Loc ( long cc, Short ip )
{	if ( !show_1st_location ) return;
	CYCLES = cc;
//	PrintLine(CStr((char*)GetRegisters()));
	PrintInfo ( ip );
}


// -----	Log first use of any instruction --------------------------------
//		called before executing instruction and after decrementing cc by 4
//		informational call only!
void Z80_1st_Instr ( long cc, Short ip )
{	if ( !show_1st_instr ) return;
	CYCLES = cc;
	PrintInfo ( ip );
}




