
/*	event polling and decoding and apple event handler

***	Copyright:	1994-2001 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	---------------------------------------------------------------------------------

	Call "InitAppleEvents()" at an early point in your application, 
	after you have checked using "Gestalt()" that apple events are available, 
	and before you poll Mac OS events for the first time.

	The four required apple events result in the following action:
	'oapp'	HandleNEW()
	'odoc'	HandleOPEN(FSSpec)
	'pdoc'	HandlePRINT(FSSpec)
	'quit' 	done := true		// check this flag frequently!

	---------------------------------------------------------------------------------

	Call "PollEvents(mask,delay,region)" whenever you want to poll Mac OS events.
	"PollEvent()" polls Mac OS events once, calls in return one of the following routines, 
	which you must supply in your main application and returns:

	HandleNEW()									(application launched)
	HandleOPEN (FSSpec)							(via AEProcessAppleEvent())
	HandlePRINT (FSSpec)						(dito)
	HandleDISK ()
	HandleDRIVER ()								(only device drivers)
	HandleAPP3 ()								(old style inter-application event)
	HandleKEYDOWN (Char)
	HandleAUTOKEY (Char)
	HandleKEYUP (Char)							(this event needs to be enabled explicitly)
	HandleMENU (menu,item)
	HandleCLOSE (WindowPtr)						(closebox top window or 'apple'+'W')
	HandleZOOM (WindowPtr)						(zoombox top window)
	HandleGROW (WindowPtr)						(growbox top window; only required if SCROLLBARS==FALSE)
	HandleINCONTENT (WindowPtr)					(content region of window at event.where (local coord.))
	HandleBUTTON(WindowPtr,ControlHandle)		(button at event.where (local coord.))
	HandleCHECKBOX(WindowPtr,ControlHandle)		(checkbox or radio button at event.where (local coord.))
	HandleSLIDEBAR(WindowPtr,ControlHandle,partCode)	(mouse down in slide bar -> track it)
	HandleACTIVATE (WindowPtr)					(window)
	HandleDEACTIVATE (WindowPtr)				(window)
	HandleUPDATE (WindowPtr)
	HandleFASTUPDATE(WindowPtr)				(redraw window->updateRgn if you can do it fast. Only required if SCROLLBARS==FALSE)
	HandleMOUSEUP ()						(this event may be disabled)
	HandleRESUME ()							(application)
	HandleSUSPEND ()						(application)
	HandleREGION()							(mouse leaves region)
	HandleNULL()							(no event available. in addition 'idle' is incremented)

	---------------------------------------------------------------------------------

	20.Nov.93	Started work on this file												KIO !
	22.Jun.94	removed bug in handling required AppleEvent 'quit'						KIO !
	09.Jul.94	event polling and dispatching now outside main() in separate function	KIO !
	14.Mar.95	PollEvents() now polls only one single event							KIO !
	28.Mar.95	PPC adaption															KIO !
	26.Mai.95	'oapp' now calls HandleNEW()											KIO !
	09.Sep.95	HandleBUTTON(), HandleCHECKBOX()										KIO !
	12.Sep.95	FindRefControl(), HandleSlidebar() and HandleGrowicon()					KIO !
	20.Apr.96	slidebar handling now optional, standard actions for cmd+Q/W/.			KIO !
	14.Dec.97	PollEvents() now returns a flag to indicate event processed or not		KIO !
	29.Dec.97	PollEvents() now always detects cmd-Q/W/. before scanning menus			KIO !
    20.Oct.01   revised for CW7Pro and ANSI C compliance								hdusel@bnv-gz.de

	---------------------------------------------------------------------------------

	Instructions for slide bars and grow icon
	-------------------------------------
	
	PollEvents() now handles slide bars and dragging of the grow box internally. It only calls
	HandleUPDATE() or HandleFASTUPDATE() requesting the application to redraw revealed
	parts of the screen. By this way it implements 'power scrolling'.
	The application does not need to draw the slide bars or the grow icon.
	The application does not need to disable or enable the slide bars if a window is dis/enabled.
	
	HandleSlidebar() and HandleGrowicon() are only called, if a window has a grow icon. 
	Windows with slide bars and without grow icon are illegal.
	HandleEvents() provides means of automatically add and remove slide bars as needed.
	If you create a new window without slide bars but with a grow icon, PollEvents() allows
	the user to shrink the window and adds slide bars as needed. The application must always 
	test for the presence of slide bars when it draws to the window!
	Unneeded slide bars are only removed if they extend to the full width or height of the window.
	If left from or above of a slide bar is a gap, the slide bars remain present, even if the window
	is full sized.
	
	If the application adds slide bars to a window manually, obey the following rules:
	• set the contrlRfCon field to 'hsb_magic' or 'vsb_magic'. These constants are defined in "pollevents.h"
	   This is the last parameter in the NewControl() call. 
	• contrlMin must be the minimum value for the window origin
	• contrlValue must be the current value of the window origin
	• contrlMax must be the maximum value for the window origin

	PollEvents() updates the contrlMin/Value/Max fields if the window is scrolled/resized.
	If the window is scrolled, the windows origin is moved accordingly.
*/

#define         OLDROUTINENAMES 1

#include		<QuickDraw.h>
#include		<AppleEvents.h>
#include		<windows.h>

#if UNIVERSAL_INTERFACES_VERSION > 0x0214
#include        <ControlDefinitions.h>
#endif

#include		"kio.h"			// kio's standard definitions
#include		"options.h"


// ==========================================================================
// ============================== PROCEDURES =================================
// ==========================================================================


/* ----	Find control with specific contrlRfCon value ------------------------------------------
		returns handle or nil if not found
*/
ControlHandle FindRefControl ( WindowPtr window, long refValue )
{	ControlHandle con;

	for ( con = (ControlHandle)((WindowRecord*)window)->controlList; con; con=(**con).nextControl )
	{	if ((**con).contrlRfCon==refValue) return con;
	};
	return nil;
}


#if !SCROLLBARS
#define	dsb(W,F)
#endif

#define	DrawSlidebars(W)	dsb(W,1)
#define	ShowSlidebars(W)	dsb(W,2)
#define	HideSlidebars(W)	dsb(W,0)

#if SCROLLBARS
#define	min_sbl 61		// min. slide bar length; at least 61 !!!
#define	HandleGROW(W)		HandleGrowicon(W)


/* ----	Scroll rectangle region of window -----------------
*/
void KIOScrollWindowRect ( WindowPtr window, Rect *box, short dx, short dy )
{	RgnHandle tempRgn,updtRgn;
	Rect wbox;

	SetPort ( window );	

	wbox = window->portRect;
	if (FindRefControl(window,vsb_magic))	{ wbox.right-=15;  wbox.top=(**FindRefControl(window,vsb_magic)).contrlRect.top+1; }
	if (FindRefControl(window,hsb_magic))	{ wbox.bottom-=15; }
	SetRect ( &wbox,  Max(wbox.left,box->left), Max(box->top,wbox.top), Min(wbox.right,box->right), Min(box->bottom,wbox.bottom) ); 

	updtRgn = NewRgn();
	tempRgn = NewRgn();
	if (tempRgn) 
	{
		Point tp={0,0}; GlobalToLocal(&tp);			
		CopyRgn ( ((WindowRecord*)window)->updateRgn, updtRgn );	// get a copy of the dirty region
		OffsetRgn(updtRgn,tp.h+dx,tp.v+dy);			// convert update region to local coordinates and scroll update region

		RectRgn ( tempRgn, &wbox );					// make a region of the scroll area
		ValidRgn(tempRgn);							// remove scroll area from update region
		SectRgn (tempRgn, updtRgn, updtRgn);		// clip to scrollbox
		ScrollRect ( &wbox, dx, dy, tempRgn);		// blit the bits ...
		InvalRgn ( tempRgn );						// and mark revealed area as dirty
		InvalRgn ( updtRgn );						// and mark als dirty again
	}
	else
	{	InvalRect ( &wbox );
	}
	if (updtRgn)	DisposeRgn ( updtRgn );
	if (tempRgn)	DisposeRgn ( tempRgn );
}


/* ----	draw growbox and slide bars ------------------------------------------
		active window -> full color
		inactive window -> dimmed
		GrafPort must be set before calling DrawSlidebars()!
		Slide bars are cleared with white color if dimmed
		Clip region is set to exclude slidebars
		flag = 0  =>  HideSlidebars()
		       = 1  =>  DrawSlidebars()
		       = 2  =>  ShowSlidebars()
*/
static void dsb ( WindowPtr window, short flag )
{	Rect box;
	ControlHandle hsb,vsb;
	RGBColor bc;
	
	hsb = FindRefControl(window,hsb_magic);
	vsb = FindRefControl(window,vsb_magic);
	ClipRect(&window->portRect); 
	if (!(hsb||vsb)) return;

	switch (flag)
	{
	case 0:	// HideSlidebars()
		GetBackColor(&bc);
		BackColor(whiteColor);
		if (hsb) HideControl(hsb); 
		if (vsb) HideControl(vsb);
		RGBBackColor(&bc);
		break;
	case 1:	// DrawSlidebars()			
		if (vsb) Draw1Control (vsb);
		if (hsb) Draw1Control (hsb);
		break;
	default:	// ShowSlidebars()
		if (vsb) ShowControl(vsb);
		if (hsb) ShowControl(hsb);
	}

	box = window->portRect;
	if (!vsb) box.top  = box.bottom-15;		// no vert. slide bar
	if (!hsb) box.left = box.right-15;		// no hori. slide bar
	ClipRect (&box);
	DrawGrowIcon (window);

	if (hsb) ValidRect ( &(**hsb).contrlRect );
	if (vsb) ValidRect ( &(**vsb).contrlRect );
	box = window->portRect; box.left=box.right-15; box.top=box.bottom-15; ValidRect ( &box );
	
	box = window->portRect; 
	if (vsb)	{ box.top=(**vsb).contrlRect.top+1; box.right-=15; }
	if (hsb)	{ box.bottom-=15; }
	ClipRect (&box);
}


/* ----	Resize window and add / remove scrollbars as required -----------------
*/
void KIOResizeWindow ( WindowPtr window, short x, short y, short w, short h )
{	Rect page;				// entire page
	Short width, height;		// of entire page
	Rect	oldbox;			// old scrollable region
	Rect	newbox;			// new scrollable region
	ControlHandle hsb,vsb;	// handles to scroll bars
	Boolean hsbf,vsbf;		// scroll bars required?
	Rect box;

// find slide bars	
	hsb = FindRefControl(window,hsb_magic);
	vsb = FindRefControl(window,vsb_magic);

// get old scroll area (excluding scroll bars and status lines at top of window)
	oldbox = window->portRect;
	if (hsb)	{ oldbox.bottom -= 15; }
	if (vsb)	{ oldbox.right-=15; oldbox.top=(**vsb).contrlRect.top+1; }

// calculate box for entire image
// note: there may be an offset between oldbox.top/left and vsb/hsb.contrlValue !!!
	page = oldbox;
	if (hsb)	
	{	page.left		+= (**hsb).contrlMin - (**hsb).contrlValue;
		page.right		+= (**hsb).contrlMax - (**hsb).contrlValue;
	}
	if (vsb)
	{	page.top		+= (**vsb).contrlMin - (**vsb).contrlValue;
		page.bottom	+= (**vsb).contrlMax - (**vsb).contrlValue;
	}
	width = page.right-page.left;
	height = page.bottom-page.top;

// which slide bars are needed?
	hsbf = hsb ? (**hsb).contrlRect.left>=window->portRect.left : false;		// don't remove hsb if anything is left of it
	vsbf = vsb ? (**vsb).contrlRect.top>=window->portRect.top : false;		// don't remove vsb if anything is above it
	hsbf |= w-vsbf*15 < width;
	vsbf |= h-hsbf*15  < height;
	hsbf |= w-vsbf*15 < width;

// size new scroll region
	w = MinMax ( min_sbl, w-vsbf*15, width );
	h  = MinMax ( min_sbl, h-hsbf*15-(vsb?(**vsb).contrlRect.top+1-window->portRect.top:0),  height );
	x = MinMax ( page.left, x, page.right-w );
	y = MinMax ( page.top,  y, page.bottom-h );
	SetRect ( &newbox, x,y, x+w,y+h );

// invalidate old slide bar boxes
	if (hsb) { box = window->portRect; box.top=box.bottom-15; InvalRect(&box); }
	if (vsb) { box = window->portRect; box.left=box.right-15; box.top=(**vsb).contrlRect.top+1; InvalRect(&box); }

// dispose off unneeded slidebars  ( should be done before calling ScrollWindowRect() )
	if (hsb && !hsbf) { DisposeControl(hsb); hsb=nil; }
	if (vsb && !vsbf) { DisposeControl(vsb); vsb=nil; }

// resize window and scroll scrollable region if needed
	if ( newbox.left!=oldbox.left || newbox.top!=oldbox.top )	// origin moved
	{	box = oldbox;
		box.right = Max ( box.right, box.left+w );
		box.bottom = Max ( box.bottom, box.top+h );
		SizeWindow ( window, box.right-box.left+vsbf*15, ((vsb)?(**vsb).contrlRect.top+1-window->portRect.top:0)+box.bottom-box.top+hsbf*15, true);
		ClipRect ( &box );
		KIOScrollWindowRect ( window, &box, oldbox.left-newbox.left, oldbox.top-newbox.top );
	}
	SizeWindow (window, w+vsbf*15, ((vsb)?(**vsb).contrlRect.top+1-window->portRect.top:0)+h+hsbf*15, true);
	
// update slide bars:
	if (hsbf)
	{	box=window->portRect;
		SetRect ( &box, hsb?(**hsb).contrlRect.left:box.left-1, box.bottom-15, box.right-14, box.bottom+1 );
		if (!hsb) 
		{	hsb = NewControl ( window, &box, "\p", true, x, page.left, page.right-w, scrollBarProc, hsb_magic );
			if (!hsb) { SysBeep(20); return; }
		}
		else
		{	(**hsb).contrlValue	+= newbox.left - oldbox.left;
			(**hsb).contrlMax	+= oldbox.right-oldbox.left-w;
			(**hsb).contrlRect = box;
		}
	}
	
	if (vsbf)
	{	box=window->portRect;
		SetRect ( &box, box.right-15, vsb?(**vsb).contrlRect.top:box.top-1, box.right+1, box.bottom-14 );
		if (!vsb) 
		{	vsb = NewControl ( window, &box, "\p", true, y, page.top, page.bottom-h, scrollBarProc, vsb_magic );
			if (!vsb) { SysBeep(20); return; }
		}
		else
		{	(**vsb).contrlValue	+= newbox.top - oldbox.top;
			(**vsb).contrlMax	+= oldbox.bottom-oldbox.top-h;
			(**vsb).contrlRect = box;
		}
	}
	
// origin of window moved ?
	x=newbox.left-oldbox.left;
	y=newbox.top-oldbox.top;
	if ( x || y )
	{	SetOrigin ( window->portRect.left+x, window->portRect.top+y );
		if (hsb) OffsetRect ( &(**hsb).contrlRect, x, y );
		if (vsb) OffsetRect ( &(**vsb).contrlRect, x, y );
		//OffsetRgn ( ((WindowRecord*)window)->updateRgn, x, y );
	}

	DrawSlidebars(window);
//	ClipRect(&newbox);
}


/* ----	Handle dragging of Grow icon ---------------------------------------------------
		The following is true for both slide bars respectively:
		•	slide bars are added if they were not present and window shrinks
			minimum window size is big enough to allow fully functional slide bars
		•	slide bars are removed if window is expanded to full size
			except: slide bars are not removed if they do not start at the left or top border of the window
		the window can only be expanded to it's maximum size as indicated by slide bars
		if window is full sized, the grow icon may be invisible but is still present
		This routine assumes, that contrlMin, contrlValue and contrlMax define the remaining scrollable distance in window pixels
		contrlMin = origin of virtual page = min. value for window origin
		contrlValue = origin of window 
		contrlMax = max. value for window origin (depends on window size!)
*/

static void HandleGrowicon ( WindowPtr window )
{	short width,height,x,y;
	Rect box,wbox;
	ControlHandle hsb,vsb;

// find slide bars	
	hsb = FindRefControl(window,hsb_magic);
	vsb = FindRefControl(window,vsb_magic);

// calculate origin and size of entire image
	wbox = window->portRect;
	x = wbox.left;
	y = wbox.top;
	width = wbox.right-wbox.left;
	height = wbox.bottom-wbox.top;
	if (hsb)	{	height -= 15;	width += (**hsb).contrlMax- (**hsb).contrlMin;	x = (**hsb).contrlValue;	}
	if (vsb)	{	width -= 15;	height += (**vsb).contrlMax-(**vsb).contrlMin;	y = (**vsb).contrlValue;	}

// calculate min/max size rectangle
	SetRect ( &box, min_sbl, min_sbl, width+16,  height+16 );
	if (hsb)	{	box.left	+= (**hsb).contrlRect.left+1-wbox.left; }
	if (vsb)	{	box.top	+= (**vsb).contrlRect.top+1-wbox.top; }

#if POWERSIZE
{	Point old,start;
	Rect oldbox;
	
	GlobalToLocal(&event.where);
	start = old = event.where;
	oldbox = window->portRect;

	while(StillDown())
	{	while ( StillDown() && event.where.h==old.h && event.where.v==old.v ) 
		{	//WaitNextEvent(0,&event,0,nil);		// this will result in background windows beeing redrawn permanently too
			GetMouse(&event.where);
		}
		old=event.where;
		if (StillDown()) 
		{	ResizeWindow ( window, x, y,  oldbox.right - window->portRect.left + old.h - start.h,  oldbox.bottom - window->portRect.top + old.v - start.v 	);
			HandleFASTUPDATE(window);
		}
	}
}
#else
{	long newsize;
	newsize = GrowWindow (window, event.where, &box); if (!newsize) return;	// track grow icon
	KIOResizeWindow ( window, x, y, newsize, newsize>>16 );					// size window
}
#endif
}


/* ----	Handle Action in Slide bar ------------------------------------------------------
		Handles clicking in arrow button, page region and thumb
		Origin of window is repositioned accordingly
		HandleFASTUPDATE() is called repeatedly to refresh revealed areas immediately (-> power scroll)
*/
static void HandleSlidebar ( WindowPtr window, ControlHandle control, short where )
{	short	delta,dx,dy;
	Boolean	hilit,hsb,vsb;
	RgnHandle	invalrgn;
	ControlHandle othercontrol;					// handle to other slidebar
	short	oldContrlValue = (**control).contrlValue;	// needed if user cancels draging thumb
	Rect	wbox = window->portRect;				// window's bounding box
	Rect	cbox  = (**control).contrlRect;			// controls bounding box
	Rect	xbox;								// controls bounding box, widened

	hilit	= where==inUpButton||where==inDownButton||where==inThumb;	// needing hilighting?
	hsb	= cbox.left<wbox.right-16;  vsb = !hsb;					// hor. scrollbar? (else: vsb)
	xbox	= cbox; InsetRect (&xbox, -16, -16);					// within this box draging thumb is 'not canceled'
	invalrgn = NewRgn(); if (!invalrgn) { SysBeep(20); return; }

	othercontrol = FindRefControl(window,hsb?vsb_magic:hsb_magic);	// search for other slide bar

	if (othercontrol||hsb) wbox.bottom-=15;						// subtract scroll bars from contents box
	if (othercontrol||vsb) { wbox.right-=15; wbox.top=(**(vsb?control:othercontrol)).contrlRect.top+1; } 

	if (hilit)	delta = where==inUpButton ? -16 : 16;				// scrolling: 16 pixel
	else	{	delta = ( hsb ? wbox.right-wbox.left : wbox.bottom-wbox.top ) * 2 / 3;
			if (where==inPageUp) delta = -delta;					// paging: 2/3 screen
		}

	do 
	{	ClipRect(&window->portRect);
		if (hilit) HiliteControl(control,where);
		do 
		{	if (where==inThumb)
			{	if (PtInRect (event.where, &xbox))
				{	if (hsb)	delta =	(**control).contrlMin + ( ((**control).contrlMax-(**control).contrlMin) * (event.where.h-cbox.left-8-16) + (cbox.right-cbox.left-16-16-16)/2 ) / (cbox.right-cbox.left-16-16-16)   - (**control).contrlValue;
					else		delta =	(**control).contrlMin + ( ((**control).contrlMax-(**control).contrlMin) * (event.where.v-cbox.top-8-16) + (cbox.bottom-cbox.top-16-16-16)/2 ) / (cbox.bottom-cbox.top-16-16-16) - (**control).contrlValue;
				}
				else			delta =	oldContrlValue- (**control).contrlValue;
			}
			delta = Min ( delta,  (**control).contrlMax-(**control).contrlValue );
			delta = Max ( delta,  (**control).contrlMin-(**control).contrlValue );
			(**control).contrlValue += delta;

			if (delta)
			{	dx = delta*hsb; dy = delta*vsb;
				SetOrigin(wbox.left+dx,wbox.top+dy);
				OffsetRect(&wbox,dx,dy);
				OffsetRect(&cbox,dx,dy);
				OffsetRect(&xbox,dx,dy);
				OffsetRect(&(**control).contrlRect,dx,dy);
				OffsetRect(&(**othercontrol).contrlRect,dx,dy);
				OffsetRgn (((WindowRecord*)window)->updateRgn, -dx, -dy);		// added: 20.apr.96 KIO !
				ClipRect(&window->portRect);
				ScrollRect (&wbox, -dx, -dy, invalrgn); 
				InvalRgn(invalrgn);
				HandleFASTUPDATE(window);
				UpdtControl(window,invalrgn);
				DrawSlidebars(window);
				WaitNextEvent(0,&event,hilit?0:10,nil);
			}
			else	WaitNextEvent(0,&event,10,nil);

			GetMouse(&event.where);
		} while ( (where==inThumb||TestControl(control,event.where)==where) && StillDown() );
		
		ClipRect(&window->portRect);
		if (hilit) HiliteControl(control,0);

		while ( TestControl(control,event.where)!=where && StillDown() )
		{	WaitNextEvent(0,&event,10,nil);
			GetMouse(&event.where);
		}
	} while ( StillDown() );
	
	DisposeRgn(invalrgn);
	DrawSlidebars(window);
}

#endif


/* ****	THE MAIN EVENT POLLING ROUTINE *********************************
		Fetches next event and calls appropriate handler
		Fetches exactly one event per call
		Detects and handles 'apple'+'Q'/'W'/'.' even if disabled or not asigned in any menu 
*/
Boolean PollEvents ( short evMask, long evDelay, RgnHandle evRegion )
{	long selection;
	WindowPtr window; short where; 
	ControlHandle control;

	if ( WaitNextEvent(evMask,&event,evDelay,evRegion) )
	{	switch (event.what)
		{
		case kHighLevelEvent:	AEProcessAppleEvent(&event); 			return true;
		case mouseUp:			HandleMOUSEUP(); 			 			return true;
		case keyUp:				HandleKEYUP (event.message);  			return true;
		case autoKey:			HandleAUTOKEY (event.message); 			return true;
		case diskEvt:			HandleDISK();  							return true;
		case driverEvt:			HandleDRIVER(); 						return true;
		case app3Evt:			HandleAPP3(); 							return true;
		case mouseDown:		
		{	where = FindWindow ( event.where, &window );
			switch (where)
			{
			case inDrag:	
				SetPort(window);
				if(window!=FrontWindow())
				{	SelectWindow (window); 
					HandleFASTUPDATE(window);
					ShowSlidebars(window);
				};
#if USE_OWN_DRAG
				HandleDRAG ( window, event.where );					return true;
#else
				DragWindow ( window, event.where, &qd.screenBits.bounds );	return true;
#endif
			case inSysWindow:
				SystemClick ( &event, window );			return true;
			case inGoAway:	
				SetPort(window);
				HandleCLOSE(window);				return true;
			case inGrow:	
				SetPort(window);
				SelectWindow (window); 
				HandleGROW(window);				return true;
			case inZoomIn:	
				SetPort(window);
				SelectWindow (window); 
				HandleZOOM(window);					return true;
			case inZoomOut:	
				SetPort(window);
				SelectWindow (window); 
				HandleZOOM(window);					return true;
			case inMenuBar:	
				selection = MenuSelect( event.where );
				if (HiWord(selection))  HandleMENU (  HiWord(selection), LoWord(selection)  );
				HiliteMenu (0);				
				return true;
			case inContent:		
				SetPort(window);
				SelectWindow (window); 
				GlobalToLocal ( &event.where );
				where = FindControl (event.where, window, &control);
				switch (where)
				{
					case 0:		HandleINCONTENT(window); 										return true;
					case inButton:	if (TrackControl (control, event.where, nil)) HandleBUTTON(window,control);		return true;
					case inCheckBox:	if (TrackControl (control, event.where, nil)) HandleCHECKBOX(window,control);	return true;
					case inThumb:
					case inUpButton:
					case inDownButton:
					case inPageUp:
					case inPageDown:
#if SCROLLBARS
						if ((**control).contrlRfCon==hsb_magic || (**control).contrlRfCon==vsb_magic) 
						{	HandleSlidebar(window,control,where);
							return true;
						}
#endif
						HandleSLIDEBAR(window,control,where);
				}
				return true;
			}
		};	return true;
		case keyDown:
			if  ( event.modifiers&cmdKey && evMask&mDownMask )
			{
				switch(event.message&0x00ff)
				{
				case 'w':	
				case 'W':	window=FrontWindow(); if (window) HandleCLOSE(window); return true;
				case 'q':
				case 'Q':	done=true;   return true;
				case '.':	return true;
				}
				selection = MenuKey(event.message&0x00ff);
				if (HiWord(selection))  
				{	HandleMENU ( HiWord(selection), LoWord(selection) );
					HiliteMenu (0);
					return true;
				}
				SysBeep(20);
				return true;
			}
			HandleKEYDOWN (event.message); 
			return true;
		case activateEvt:	
			window = (WindowPtr)event.message;
			SetPort (window);
			if (event.modifiers&activeFlag)	
			{	HandleACTIVATE (window); 			// window activated  
				ShowSlidebars(window);
			}
			else
			{	HandleDEACTIVATE (window);			// window deactivated
				HideSlidebars (window);
			}
			return true;
		case updateEvt:	
			window 	= (WindowPtr)event.message;
			SetPort		(window);
			BeginUpdate	(window);
			HandleUPDATE	(window);
			UpdtControl	(window,window->visRgn);
			DrawSlidebars	(window);
			EndUpdate   	(window); 
			return true;
		case osEvt:
			if ((event.message & 0xFF000000)==0xFA000000)
				HandleREGION();						// mouse leaves eventRegion
			if (event.message & 0x01000000 ) 
			{	SetCursor(&qd.arrow);
				if ( event.message & 0x00000001 ) 
				{	HandleRESUME(); 				// application re-activated 
					window = FrontWindow();
					if (window)
					{	SetPort(window);
						HandleACTIVATE(window);
						ShowSlidebars(window);
					}
				}
				else	
				{	window = FrontWindow(); 
					if (window)
					{	SetPort(window);
						HandleDEACTIVATE(window);
						HideSlidebars(window);
					}
					HandleSUSPEND(); 				// application de-activated
				}
			}
			return true;
		};
		return true;									// unknown event
	}

	HandleNULL();
	return false;
}


/*	recursively open files in folder
	--------------------------------

	if your application needs support for processing all files in a folder, enable this option in "option.h"
	and add file types 'fold' and 'disk' to the applications BNDL resource.
	
	since this may result in a huge amount of HandleOPEN() calls, 
	queuing the calls may be a good idea for the future...
*/

#if OPEN_FOLDERS

void HandleFOLDER ( FSSpec& fs )
{
	CInfoPBRec			pb;						// union { HFileinfo;DirInfo }

	pb.hFileInfo.ioNamePtr		= fs.name;		// in: full or partial path; out: filename (Str31)
	pb.hFileInfo.ioVRefNum		= fs.vRefNum;	// in: volume, drive or working dir ID
	pb.hFileInfo.ioFDirIndex	= 0;			// in: query 1 item
	pb.hFileInfo.ioDirID		= fs.parID;		// in: hard ID of dir to search; out: hard ID of file or dir
	
	err = PBGetCatInfo ( &pb, false );
	if (err) return;

// handle file
	if ( (pb.hFileInfo.ioFlAttrib & 0x10) == 0 )
	{
		HandleOPEN(fs);
		return;
	}
	
// handle directory
	long	dirIDToSearch = pb.hFileInfo.ioDirID;

	for ( int index=1; !(done||abo); index++ )
	{
		pb.hFileInfo.ioNamePtr	 = fs.name;			// in: ignored; out: file or dir name
		pb.hFileInfo.ioVRefNum	 = fs.vRefNum;		// volume
		pb.hFileInfo.ioDirID	 = dirIDToSearch;	// hard ID of dir
		pb.hFileInfo.ioFDirIndex = index;			// index to querry

		err= PBGetCatInfo ( &pb, false );
		if (err) return;

		FSSpec	nfs;
		FSMakeFSSpec ( fs.vRefNum, dirIDToSearch, pb.hFileInfo.ioNamePtr, &nfs );
		HandleFOLDER ( nfs );
	}
}

#endif






