/*	Profiler functions

***	Copyright:	1995-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	28.Jan.95 	Implemented command profiling					KIO !
	19.Mar.95:	Extracted this file from application.c
    20.Oct.01   revised for CW7Pro and ANSI C compliance									hdusel@bnv-gz.de
*/

#include	<string.h>

#include	"kio.h"
#include	"z80.options"
#include	"z80.h"
#include	"profiler.h"
#include	"application.h"
#include	"messages.h"
#include	"menue.h"
#include	"disassembler.h"
#include	"strings.h"


#define MemSet(a,b,c) memset(b,a,c)
#define MemCopy(a,b,c) memcpy(b,a,c)
#define PrintLine(a) PrintLine(PStr(a))
#define PrintStr(a) PrintStr(PStr(a))

#if CMD_PROFILE

	Boolean	count_instr	= false;

	long	*cnt_xx = nil;		// array: count use of normal instructions
	long	*cnt_cb = nil;		// array: count use of CB instructions
	long	*cnt_ed = nil;		// array: count use of ED instructions
	long	*cnt_xy = nil;		// array: count use of ix and iy instructions
	long	*cnt_xycb = nil;	// array: count use of ix/CB and iy/CB instructions

#endif


#if PC_PROFILE

	Boolean	count_pc = false;

	long	*cnt_pc	= nil;		// array: count instruction executions per location in RAM

#endif


/* ----	Clear z80 instruction profile -----------------------------------------
		allocate arrays for profiling and clear them
		counting starts from -1 to speed up 68k coded z80 engines
*/
void ClearInstrProfiles ( )
{	
#if CMD_PROFILE
	if (!cnt_xx)	cnt_xx = (long*)NewPtr ( 5*0x0400 );
	if (!cnt_xx)	abort_oomem ( "\pClearInstrProfiles()" );
	cnt_cb	= cnt_xx+0x100;
	cnt_ed	= cnt_xx+0x200;
	cnt_xy	= cnt_xx+0x300;
	cnt_xycb= cnt_xx+0x400;
	MemSet ( 0xff, cnt_xx, 5*0x0400 );
#endif
}


/* ----	Clear executed location profile -----------------------------------------
		allocate array for profiling and clear it
		counting starts from -1 to speed up 68k coded z80 engines
*/
void ClearLocProfiles ( void )
{	OSErr err = ok; 
	static Handle h = nil;

#if PC_PROFILE
	if (!h)			h = TempNewHandle ( 0x00040000, &err );
	if (!h||err)	abort_oomem ( "\pClearLocProfiles()" );
	HLock(h);
	cnt_pc = *(long**)h;
	MemSet ( 0xff, cnt_pc, 0x00040000 );
#endif
}


// ----	Show 'number of executions per location' profile -------------------------------------
//		'short'
void ShowPcProfile ( void)
{
#if PC_PROFILE
	long i,j,k,m;

	if (!log_file) HandleMENU(mOptions,iLogToFile);

	PrintNL();
	PrintNL();
	PrintLine ( "\pNumber of executions per location [granularity=16]" );
	PrintLine ( "\p--------------------------------------------------" );
	PrintNL();
	for ( i=0; i<0x00010000; i+=256 )
	{	PrintHex4(i);
		PrintChar(':');
		for ( j=i; j<i+256; j+=16 )
		{	for ( k=j, m=0; k<j+16; k++ ) m = max ( m, cnt_pc[k]+1 );
			if (m>=0x00010000)	PrintStr("\p****");
			else if (m)				PrintHex4(m);
			else					PrintStr("\p----");
		};
		PrintNL();
	};
	PrintNL();
#endif
}


// -----	Show Executed Locations ---------------------------------------------
//		• = instruction byte
//		o = operand byte

void ShowExecutedLocations ( void )
{
#if PC_PROFILE
	long i,j,k,size;
	Char *c;
	Str255 s;
	
	if (!log_file) HandleMENU(mOptions,iLogToFile);

	PrintNL();
	PrintNL();
	PrintLine ( "\pMap of executed locations [•/o = opcode/operand]" );
	PrintLine ( "\p------------------------------------------------" );
	PrintNL();
	*s = 64; MemSet ('-',s+1,80);
	for ( i=0; i<0x00010000; i+=64 )
	{	c = s+1; MemCopy ( c+64, c, 4 );
		MemSet ( '-', c+4, 64 );
		for ( j=i; j<i+64; j++ )
		{	if ( cnt_pc[j]+1 )
			{	size = OpcodeLength ( Peek(j),Peek(j+1) );
				*c = '•';							// mark instruction
				for (k=1; k<size;k++) c[k] = 'o';		// mark arguments
			};
			c++;
		};
		PrintHex4(i);
		PrintChar(':');
		c=s;    *c=16; PrintStr ( c );
		c+=16;*c=16; PrintStr ( c );
		c+=16;*c=16; PrintStr ( c );
		c+=16;*c=16; PrintLine ( c );
	};
	PrintNL();
#endif
}


// ----	Show Used Instructions ----------------------------------------------
void ShowUsedCmds ( void )
{
#if CMD_PROFILE
	short i,j,k,l;
	long *p;
	Char *c;
	Str255 s;

	if (!log_file) HandleMENU(mOptions,iLogToFile);

	p=cnt_xx;
	for (i=0;i<=4;i++)
	{	PrintNL();
		PrintNL();
		switch(i)
		{	case 0: 	PrintLine("\pInstructions $xx" ); 				break;
			case 1: 	PrintLine("\pInstructions $CBxx" );				break;
			case 2: 	PrintLine("\pInstructions $EDxx" );				break;
			case 3: 	PrintLine("\pIndex register instructions $xx" );	break;
			case 4: 	PrintLine("\pIndex register instructions $CBxx");	break;
		};
		PrintLine ( "\p----------------------------------------" );
		PrintNL();
		for ( j=0; j<256; j+=32 )
		{	PrintHex2(j);PrintChar(':');
			c = s; *c++ = 35;
			for ( k=4; k; k-- )
			{	for ( l=8; l; l-- ) *c++ = (*p++ +1) ? '•' : '-';
				*c++ = ' ';
			};
			PrintLine ( s );
		};
	};
	PrintNL();	
#endif
}


// ----	Show instruction profile --------------------------------------------
void ShowCmdProfile ( void )
{
#if CMD_PROFILE
	short i,j;
	long *p,cnt;
	string s;
			
	if (!log_file) HandleMENU(mOptions,iLogToFile);

	p=cnt_xx;
	for ( i=0; i<=4;i ++ )
	{
		PrintNL();
		PrintNL();
		switch(i)
		{	
		case 0: PrintLine("\pNormal Instructions"); 	break;
		case 1: PrintLine("\pCB Instructions");			break;
		case 2: PrintLine("\pED Instructions");			break;
		case 3: PrintLine("\pIX/IY Instructions");		break;
		case 4: PrintLine("\pIX/IY CB Instructions");	break;
		};
		
		PrintLine ( "\p---------------------" );
		PrintNL();
		
		for ( j=0; j<256; j++ )
		{	if ( (cnt=(*p++ +1)) != 0 )
			{	
				PrintHex2 ( j ); PrintChar ( ':' );

				#undef PrintLine
				#undef PrintStr
				
				s = CatStr(PStr("\p        "),NumStr(cnt));
				PrintStr(RightStr(s,s.len-8));
				
				switch (i)
				{	
					case 0:	PrintLine ( Mnemonic(j)     ); break;
					case 1:	PrintLine ( MnemonicCB(j)   ); break;
					case 2:	PrintLine ( MnemonicED(j)   ); break;
					case 3:	PrintLine ( MnemonicIX(j)   ); break;
					case 4:	PrintLine ( MnemonicIXCB(j) ); break;
				};
			};
		};
	};
	PrintNL();	
#endif
}


