
/*
***	Copyright:	1994-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

/* ---- String-Verwaltung für das FORCE System -----------------------------
	12.apr.94:	started word; (c) by KIO !
	22.jun.94:	SubStr(): *destStr and targetStr now may be the same	kio !
	06.nov.94:	PoolStr() entfernt. Sagte zu oft 'ja' ...				kio !
	30.jul.95:	revised													kio !
	07.aug.95:	revised													kio !
	20.apr.96:	made usable for other apps								kio !
	27.apr.96:	quick strings											kio !
	17.aug.96:	re-introduced support for FORCE system					kio !
    20.Oct.01   revised for CW7Pro and ANSI C compliance					hdusel@bnv-gz.de
*/

#include	<stdio.h>
#include	<fp.h>
//#include	<math.h>
#include	<string.h>


#include	"kio.h"
#include	"options.h"
#include	"strings.h"


#define MemCopy(a,b,c) memcpy(b,a,c)


typedef	struct slot slot;
struct slot { string* rptr; slot* next; Char text[]; };

//const string Str00 = { 0,0 }; 

string	freeString;	// für freie slots (freeString.text zeigt auf keinen Fall auf freien slot.text)
					// für letzten slot (freeString.text zeigt auf pool_end->text und freeString.len = 0)
string	nullString;	// Prototyp für leeren String ( nullString.len = 0 und nullString.addr is nicht off-memory)
slot	nullSlot;	// zugehöriger dummy-slot
slot*	pool_end;	// end of pool == last slot
slot*	pool_mid;	// pool-mitte ca.
slot*	string_pool;// start of string pool == first slot
slot*	umlptr;		// umlaufender suchzeiger

Char	c2s[256];	// für CharStr()


// ----	quick shot strings: valid until reused !!! ----------------------
#define	QS_MAX			0x10			// how many quick-shot string descriptors?  ->  must be 2^n
Short	QS_IDX;							// round-robbin allocation index
string	QS[QS_MAX];



// ================================================================================
//								P R O C E D U R E S
// ================================================================================

#if 0

#define	CHECKSTRINGPOOL	CheckStringPool(__LINE__)

static CheckStringPool(long n)
{	slot* here;
	Char x[12],y[12];
	Boolean uml=false;
	long i=99999;
	
	NumToString (n,y);

	here = string_pool;
	while (i && here!=pool_end)
	{
		if (here->next<string_pool || here->next>pool_end) 
		{	NumToString ((long)here->next, x);
			Do_Abort("\pCheckStringPool()","\p#1",y)
		}
		if (here==(slot*)umlptr) uml=true;
		here=here->next;
		i--;
	}
	if (pool_end->next!=string_pool) Do_Abort("\pCheckStringPool()","\p#2",y)
	if (!i) Do_Abort("\pCheckStringPool()","\p#3",y)
	if (!uml) Do_Abort("\pCheckStringPool()","\p#4",y)
}

#else

#define	CHECKSTRINGPOOL

#endif



/* ----	Initialize string handler -------------------------------- 18.08.96 --------
	FORCE system:
		Init FORCE system memory management
		allocate string pool in FORCE memory
*/
void InitStrings (long poolsize)
{	long i;

	string_pool = (slot*)NewPtr(poolsize);
	if (!string_pool) abort_oomem("\pInitStrings()");

	pool_end = (slot*)((long)string_pool+poolsize-sizeof(slot));
	pool_mid = string_pool + (pool_end-string_pool)/4*3;		// high water mark

	string_pool->next = pool_end;
	string_pool->rptr = &freeString;	// -> free
	pool_end->next = string_pool;
	pool_end->rptr = &freeString;	// -> used !!!

// Prototyp für leeren String initialisieren
	nullString.text	= nullSlot.text;// wir nehmen einen eigenen slot, falls jemand den rptr manipuliert; z.B. MoveMaster()
	nullString.len	= 0;			// len := 0
	nullSlot.rptr	= &nullString;	// String-Prototyp selbst ist Master (hauptsache in-memory...)

// Multifunktions-Stringdescriptor initialisieren:
// korrekten string descriptor für (slot)*pool_end bilden, damit dieser als 'belegt' gilt
// sicherstellen, daß freeString.text nicht zufällig auf einen freien slot.text im pool zeigt
	freeString.text	= pool_end->text;
	freeString.len	= 0;
	
// Umlauf-Suchzeiger initialisieren
	umlptr = string_pool;

// c2s[] initialisieren
	for (i=0; i<=255; i++) c2s[i]=i;
	
	CHECKSTRINGPOOL;
}


// ----	clear temporary string descriptors ----------------------------------------
void ClearQStrings(void)
{	string *p;

	p = QS+QS_MAX;
	while (p>QS) *--p=nullString;
}


// ----	Seach free slot in pool ---------------------------------- 27.04.96 -------
//		checks for length <= 0
//		returned string is TEMPORARY, MASTER (except: empty string is CONSTANT)

string AllocStr ( long len )
{	long i, n; string *s=NULL;
	slot *here,*next;

	if (len<=0) return nullString;							// empty string ?

	s = QS + (QS_IDX++ & (QS_MAX-1));						// get next ad-hoc string

	n = len+sizeof(slot);
	here=(slot*)umlptr;
	for (i=99999;i;i--)
	{
		if ( here->rptr->text == here->text )				// slot used
		{	here = here->next;
			continue;
		};

		next = here->next; 									// melt gaps
		while ( next->rptr->text != next->text ) 
		{	here->next = next = next->next;
		}
		
		if ( ((long)next)-((long)here) < n )				// gap too small 
		{	here = next->next;
			continue;
		}

		if ( ((long)next)-((long)here) >= n+sizeof(slot) )	// split gap
		{	next = (slot*)((long)here+n);
			next->next=here->next;
	  		next->rptr=&freeString;	// mark free
	 		here->next=next;
	  	}

	  	here->rptr=s;
	  	s->text=here->text;
	  	s->len=len;
	  	if (next>=pool_mid) umlptr=string_pool; else umlptr=next;	// update umlptr & condense
	  	return *s;

	};

	do_abort ( "\pString pool overflow", "\pAllocStr()", 0 );
	return *s;
}


// ----	copy string into pool -------------------------------- 27.04.96 ------
//		returned string is TEMPORARY, MASTER

string PoolString ( string q )
{	string s;

	CHECKSTRINGPOOL;
	s = AllocStr ( q.len );						// tests for q.len<=0
	MemCopy   ( q.text, s.text, s.len );			// guaranteed: s->len>=0
	return s;
	CHECKSTRINGPOOL;
}


// ----	Move a string --------------------------------------- 27.04.96 ------
//		source string must be MASTER
//		target becomes new MASTER
//
//		caution: do not use on SLAVE strings or CONSTANT strings!
//		use MoveMaster() to move a string variable in memory or to assign a temporary string to a string variable
//		to assign a slave string or a constant string to a variable use MoveMaster(&s,PoolString(q))
//		to move a constant string or a slave string just assign it: s=q
//		to make a copy of a variable use MoveMaster(&s,PoolString(q))

void MoveMaster ( string *s, string q )
{
	CHECKSTRINGPOOL;
	*s = q;
	if (q.text>=string_pool->text&&q.text<=pool_end->text)		// security
		((slot*)(q.text))[-1].rptr = s;
	CHECKSTRINGPOOL;
}


// ----	Prüfe, ob String leer ist -------------------------- 10.4.94 ------
//		Ein String gilt als leer, wenn er nur noch white spaces enthält.

Boolean EmptyStr ( string s )
{	Char *z; long i;

	CHECKSTRINGPOOL;
	z = s.text;
	i = s.len;
	while ( i>0 && *z++<=32 ) i--;
	return(i<=0);
}


// ----	Prüfe, ob zwei Strings gleich sind ----------------- 20.4.96 ------

Boolean SameStr ( string s, string t )
{	Char *p,*q;

	CHECKSTRINGPOOL;
	if (s.len!=t.len) return false;
	p=s.text+s.len;
	q=t.text+t.len;
	while (p>s.text) if (*--p!=*--q) return false;
	return true;
}


// ----- UP: akkumuliere Zahl aus string -----------------------------------

static long read_number ( Char *p, long *len, long base, long *err )
{	Long zahl, maxz; Char c;
	if (base<2||base>16) { *err=-1; return(0); };
	*err=nan;
	maxz = (Long)0xffffffff/(Long)base;
	for ( zahl=0; *len; (*len)-- )
	{	c=*p++;
		if (c>'9')  
		{ 	if (c>='a') c-=0x20;
			if (c<'A') return(zahl); 
			c-=7; 
		};
		c-='0'; if (c>=base||c<0) return(zahl);
		if (zahl>=maxz)
		{ 	if (zahl>maxz)  { *err=toobig; return(zahl); };
			zahl = zahl*base + c;
			if (zahl<base)  { *err=toobig; return(zahl); };
			continue;
		};
		zahl = zahl*base + c;	
		*err=ok;
	};
	return(zahl);
}

// ----- konvertiere string -> Integer -------------------------------------
// "±decimal", "±$hex", "±%bin", "'text'"

long ValStr ( string s, long *err )
{	Boolean neg; long result,len, base;
	Char *z,*e;

	CHECKSTRINGPOOL;
	s=QClipStr(s); 
	*err=0;
	z=s.text;			 
	e=z+s.len;		
	
	if ( *z=='\'' && *(e-1)=='\'' )
	{	switch (s.len)
		{	
		case 3: return ( *(z+1) );			//	'x'	unsigned char
	 	case 4: return ( *(Short*)(z+1) );	// 	'xx'	unsigned short
	  	case 6: return ( *(long*)(z+1) );	//	'xxxx'
		};
		*err=nan; return 0;
	};
		
	neg=false; if (*z=='-' || *z=='+') neg = *z++=='-';
	base=10; if (*z=='$') { base=16; z++; } else if (*z=='%') { base=2; z++; }
	len=e-z; if (len<=0) { *err=nan; return 0; };

	result = read_number ( z, &len, base, err ); if (*err) return(result);
	if (len) { *err=nan; return(result); };
	if (neg)  { if (result<0) *err=toobig; return(-result); };
	return(result);
}


#if !GENERATINGPOWERPC || GENERATINGPOWERPC // ##################################################


// ----- konvertiere string -> Float ---------------------- 11.4.94 --------
// "±dec.dec^dec"

float FValStr ( string s, long *err )	
{	long count, items;
	float zahl;
	char format[]="%88f%n";

	CHECKSTRINGPOOL;
	*err=ok;
	s=QClipStr(s); 
	zahl=0;
	if (s.len<=0) { *err=-1; return(zahl); }
	format[1]=48+(s.len)/10;
	format[2]=48+(s.len)%10;
	items = sscanf ( (char*)s.text, format, &zahl, &count);
	if (items!=1) *err=nan; 
	if (count!=s.len) *err=nan; 
	return(zahl);
}


// ----- konvertiere string -> Double ---------------------- 11.4.94 -------
// "±dec.dec^dec"

double DValStr ( string s, long *err)	
{	long count, items;
	double zahl;
	char format[]="%88lf%n";

	CHECKSTRINGPOOL;
	*err=ok;
	s=QClipStr(s); 
	zahl=0;
	if (s.len<=0) { *err=-1; return(zahl); }
	format[1]=48+(s.len)/10;
	format[2]=48+(s.len)%10;
	items = sscanf ( (char*)s.text, format, &zahl, &count);
	if (items!=1) *err=nan; 
	if (count!=s.len) *err=nan; 
	return(zahl);
}


#else // ###################################################################

#define	PLUS	0
#define	MINUS	1

static decimal StringToDecimal ( string s, long *err )
{	decimal result;
	Char	*p,*e,*z,*x;
	Boolean	vz_exp;
	string	mant, exp;

	*err = ok;
	 
	result.sgn = PLUS;
	result.exp = 0;
	result.sig.length = 0;
	
	mant = s = QClipStr(s);
	p = s.text; e = p+s.len;
	while ( p<e && *p!='^' && *p!='E' && *p!='e' ) p++;	// suche exponententrenner: '^'

// exponent
	if (p<e)
	{
		exp  = QSubStr ( s, p+1-s.text, s.len );
		mant = QSubStr ( s, 0, p-s.text );
		
		p = exp.text; e = p+exp.len;

		vz_exp = PLUS;
		if (*p=='+'||*p=='-') if (*p++=='-') vz_exp=MINUS;
		if (p>=e) *err=notanumber;

		while (p<e && *p>='0' && *p<='9' && result.exp<1000) 
			result.exp = result.exp*10+(*p++-'0');
		if (vz_exp) result.exp = -result.exp;
		if (p<e) *err=notanumber;
	}

// mantisse
	p = mant.text; e = p+mant.len;
	z = result.sig.text; x = z+SIGDIGLEN;

	if (*p=='+'||*p=='-') if (*p++=='-') result.sgn=MINUS;
	
	while (p<e && *p>='0' && *p<='9') 
	{	if (z<x) *z++ = *p++; else { p++; result.exp++; }	}
	
	if (p<e && *p=='.')
	{	p++;
		while (p<e && *p>='0' && *p<='9') 
		{	if (z<x) { *z++ = *p++; result.exp--; } else p++;	}
	}

	result.sig.length = z-result.sig.text;
	if (result.sig.length<=0 || p<e) *err=notanumber;

	while (result.sig.length && result.sig.text[0]=='0') 
		MemMove(result.sig.text+1,result.sig.text,--result.sig.length);

	return result;	
}


// ----- konvertiere string -> Float ---------------------- 17.8.96 --------
// "±dec.dec^dec"

// ••• overflow check ••• isnormal/isfinite/isnan/fpclassify

float FValStr ( string s, long *err )	
{	decimal d;

	CHECKSTRINGPOOL;
	d = StringToDecimal ( s, err );
	if (!*err) return dec2f(&d); else return NAN;
}


// ----- konvertiere string -> Double -------------------- 17.8.96 --------
// "±dec.dec^dec"

// ••• overflow check ••• isnormal/isfinite/isnan/fpclassify

double DValStr ( string s, long *err )	
{	decimal d;

	CHECKSTRINGPOOL;
	d = StringToDecimal ( s, err );
	if (!*err) return dec2num(&d); else return NAN;
}


#endif // ######################################################################



// ----	convert char to string ---------------------------- 30.7.95 ------
//		return string is CONSTANT

string CharStr ( Char c )
{	string s;

	CHECKSTRINGPOOL;
	s.text = c2s+c;
	s.len = 1;
	return s;
}


// ----	convert pascal-type string to string --------------- 30.7.95 ------
//		return string is CONSTANT (assuming the pascal-type string is constant)

string PStr ( const Str255 p )
{	string s;

	CHECKSTRINGPOOL;
	s.text=(Char*)p+1;
	s.len=*p;
	return s;
}


// ----	convert c-type string to string ----------------------- 27.4.96 ------
//		return string is CONSTANT (assuming the c-style string is constant)

string CStr ( char *c )
{	string s;

	CHECKSTRINGPOOL;
	s.text = (Char*)c;
	while(*c++) {};
	s.len = (Char*)c-s.text-1;
	return s;
}


// ----	Leerstring ------------------------------------------ 07.8.95 -----
//		Rückgabestring ist TEMPORARY, MASTER

string SpaceStr ( long n )
{	string s;
	
	CHECKSTRINGPOOL;
	s = AllocStr(n);					// prüft ob n<=0
	memset ( s.text, ' ', s.len );
	CHECKSTRINGPOOL;
	return s;
}


// ----	Convert a string back to pascal format -------------- 07.5.96 ------
//		Rückgabestring ist TEMPORARY, MASTER

Char* PascalStr ( string s )
{
	s.len=minmax(0,s.len,255);
	return CatStr(CharStr(s.len),s).text;
}


// ----	Hänge String an String an -------------------------- 07.08.95 ------
//		Rückgabestring ist MASTER

void AddString ( string *a, string b )
{	long n,m; string s;

	CHECKSTRINGPOOL;
	n=max(0,a->len);
	m=max(0,b.len);

	s = AllocStr( n+m );
	MemCopy ( a->text, s.text,   n );
	MemCopy ( b.text,  s.text+n, m );
	MoveMaster(a,s);
	CHECKSTRINGPOOL;
}


// ----	Hänge String an String an -------------------------- 27.04.96 ------
//		Rückgabestring ist TEMPORARY, MASTER

string CatStr ( string a, string b )
{	long n,m; string s;
	
	CHECKSTRINGPOOL;
	n=max(0,a.len);
	m=max(0,b.len);

	s = AllocStr( n+m );
	MemCopy ( a.text, s.text,   n );
	MemCopy ( b.text, s.text+n, m );

	CHECKSTRINGPOOL;
	return s;
}


// ----	Bilde Substring ----------------------------------- 07.08.95 ------
//		left bounded:  [a ... [e
//		char numbers count from 0 !!!
//		Rückgabestring:	MASTER/SLAVE -> SLAVE
//						CONST		 -> CONST

string QSubStr (string s, long a, long e)
{
	CHECKSTRINGPOOL;
	if (a<0) a=0;
	if (e>s.len) e=s.len;
	s.text += a;
	s.len   = e-a;
	return s;
}


// ----	Bilde Substring ----------------------------------- 27.04.96 ------
//		Rückgabestring:	MASTER/SLAVE -> SLAVE
//						CONST		 -> CONST

string QLeftStr (string s, long n)
{
	CHECKSTRINGPOOL;
	if (n<=0) return nullString;
	if (n<s.len) s.len=n;
	return s;
}


// ----	Bilde Substring ----------------------------------- 27.04.96 ------
//		Rückgabestring:	MASTER/SLAVE -> SLAVE
//						CONST		 -> CONST

string QRightStr (string s, long n)
{
	return QSubStr(s,s.len-n,s.len);
}


// ----	clip white space at both ends ---------------------- 07.08.95 ------
//		Rückgabestring:	MASTER/SLAVE -> SLAVE
//						CONST		 -> CONST

string QClipStr ( string s )
{	Char *a,*e;

	CHECKSTRINGPOOL;
	for ( a=s.text; *(a++)<=32; ) ; a--;
	for ( e=s.text+s.len; *(--e)<=32; ) ; e++;
	
	s.text = a;
	s.len = max(0,e-a);
	return s;
}


// ----	konvertiere Int -> Dezimalstring --------------------- 27.04.96 ----
//		signed, auto-sized
//		Rückgabestring ist TEMPORARY, MASTER

string NumStr ( long n )
{	Char x[12];
	
	NumToString ( n, x );
	return PString(x);
}


// ----	konvertiere Int -> Hexstring ------------------------ 27.04.96 -----
//		unsigned, masked to fit
//		Rückgabestring ist TEMPORARY, MASTER

string HexStr ( long n, long digits )
{	Char *z; long m; string s;

	CHECKSTRINGPOOL;
	if (digits<=0) return nullString;
	if (digits>8) digits=8;

	s = AllocStr(digits);
	for ( z=s.text+digits; digits; digits-- )
	{ m=n&0x0f; *--z = (m>9) ? m+55 : m+48; n>>=4; }
	CHECKSTRINGPOOL;
	return s;
}


// ----	konvertiere Int -> Binstring ------------------------ 27.04.96 -----
//		unsigned, masked to fit
//		Rückgabestring ist TEMPORARY, MASTER

string BinStr ( long n, long digits )
{	Char *z; string s;

	CHECKSTRINGPOOL;
	if (digits<=0) return nullString;
	if (digits>32) digits=32;

	s = AllocStr(digits);
	for ( z=s.text+digits; digits; digits-- )
	{ *--z = '0'+(n&1); n>>=1; }
	CHECKSTRINGPOOL;
	return s;
}


// ----	konvertiere Double -> Dezimalstring -------------- 27.04.96 --------
//		Rückgabestring ist TEMPORARY, MASTER

string DNumStr ( double d, char *format )
{	char t[128];

	*t = sprintf ( t+1, format, d );
	if (*t<=0) return PString("\p??.????");
	
	return PString((Char*)t);
}


// ----	Replace any occurance of a character by another ------------ 28.04.96 -------
//		Rückgabestring ist TEMPORARY, MASTER
string ReplaceStr ( string s, Char old, Char neu )
{	Char *p;

	s = PoolString(s);
	for ( p=s.text+s.len; --p>=s.text; ) if (*p==old) *p=neu;
	CHECKSTRINGPOOL;
	return s;
}


// ----	Convert a string to all lower case -------------------------- 28.04.96 -------
//		Rückgabestring ist TEMPORARY, MASTER
string LowerStr ( string s )
{	
	s = PoolString(s);
	LowerText ( (char*)s.text, s.len );
	CHECKSTRINGPOOL;
	return s;
}


// ----	Convert a string to all upper case ------------------------- 28.04.96 --------
//		Rückgabestring ist TEMPORARY, MASTER
string UpperStr ( string s )
{	
	s = PoolString(s);
	UpperText ( (char*)s.text, s.len );
	CHECKSTRINGPOOL;
	return s;
}


// ----	skip white spaces in string -------------------------------- 28.04.96 --------
//		idx points to the next word or to the end of the string
short SkipSpace ( string s, short i )
{
	if (i<0) i=0;
	while ( i<s.len && s.text[i]<=' ') i++;
	return i;
}


// ----	find white space in string ------------------------------------ 28.04.96 --------
//		idx points behind the actual word
short SkipWord ( string s, short i )
{
	if (i<0) i=0;
	while ( i<s.len && s.text[i]>' ') i++;
	return i;
}


// ----	find next occurance of a special character in string --------- 28.04.96 --------
//		idx points to the found char or behind string
short FindChar ( string s, Char c, short i )
{
	if (i<0) i=0;
	while ( i<s.len && s.text[i]!=c) i++;
	return i;
}


/* ====	STANDARD DIALOG BOXES ===========================================================
//
short DoAlert ( short n, string s, string t, string u, string v )
{	
	return DoAlert ( n, s.len?PascalStr(s):0, t.len?PascalStr(t):0, u.len?PascalStr(u):0, v.len?PascalStr(v):0 );
}
*/

