
/*	Main Z80 engine calling loop
	and interrupt handler

***	Copyright:	1994-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	24.Jan.95 	Extracted this file from "application.c"			KIO !
	12.Mar.95	Do_Cycles()											KIO !
	18.Mar.95	Merged with interrupts.c, improved Do_Cycles()		KIO !
	28.Mar.95	68k glue for ticker for PPC version					KIO !
	23.Apr.95	now C coded PPC ticker routine						KIO !
	26.Dec.99	made it compile again after messing up my libs...	kio !
    20.Oct.01   revised for CW7Pro and ANSI C compliance			hdusel@bnv-gz.de
*/

#include	<timer.h>
#include	<power.h>
#include	<windows.h>
#include	<string.h>

#include	"kio.h"
#include	"z80.options"
#include	"z80.h"
#include	"z80 main.h"
#include	"application.h"
#include	"messages.h"
#include	"keyboard.h"
#include	"menue.h"
#include	"load+save.h"
#include	"preferences.h"
#include	"display.h"
#include	"sound.h"
#include	"disassembler.h"
#include	"debugger.h"
#include	"strings.h"
#include	"display.h"
#include	"terminal.h"
#include	"movie.h"


// *****************************************************
//	Handling the ticker routine
// *****************************************************


//#if GENERATINGPOWERPC
#pragma options align=mac68k
//#endif

struct TMBlock	
{	QElemPtr		qLink;			//	(( TMTask ))	Queue link to next element
	short			qType;			//	(( TMTask ))	Queue element type
	TimerUPP		tmAddr;			//	(( TMTask ))	Address of pascal-type procedure
	long			tmCount;		//	(( TMTask ))	'unused' msecs: set by RmvTime()
	long			tmWakeUp;		//	InsXTime: 'last' trigger time to generate drift-free ticks
	long			tmReserved;		//	InsXTime: Apple reserved
	long	 		*frames;		//	***	 pointer -> frame counter
};
typedef struct TMBlock TMBlock;

//#if GENERATINGPOWERPC
#pragma options align=reset
//#endif

#if GENERATINGCFM
static	pascal void	TickerRoutine ( TMBlock* tblk );
#else
pascal	TMBlock*	GetTMBlock (void)  =  0x2E89;	// MOVE.L A1,(SP)
static	void		TickerRoutine (void);
#endif

static	TMBlock		tickerblock;			//	queue element for the MacOS timer
static	long		disp_magic = 0;			//	to prevent multiple initializations (security only)


// ----	The Ticker Routine !!! ----------------------- 4.Jun.95 --------------
//		ATTENTION: This routine doesn't fetch the A5 register;
//		therefore it cannot access global variables directly!

#if GENERATINGCFM
static	pascal  void  TickerRoutine(register TMBlock* tblk)  {	
#else
static	void  TickerRoutine(void)  {  register TMBlock *tblk = GetTMBlock();
#endif
	PrimeTime ( (QElem*)tblk, 20 );			// ticker interval = 20/1000s
	(*tblk->frames)++;						// increment frame count
}
#if 0
}		// just for the MW editor ...
#endif


// ----	Install ticker ---------------------------------- 4.Jun.95 --------------
void InstallTicker (void)
{	if  ( disp_magic ) return;						// already running
	tickerblock.tmWakeUp		= 0;
	tickerblock.tmReserved		= 0;
	tickerblock.tmAddr			= NewTimerProc(TickerRoutine);
	tickerblock.frames 			= &frames;
	InsXTime ( (QElem*)&tickerblock );
	PrimeTime ( (QElem*)&tickerblock, 200 );		// startup intervall = 200/1000s
	disp_magic = true;
}


// ----	Remove Ticker ----------------------------------- 26.6.94 ------------
//		*MUST* be performed before application exits !!!
void RemoveTicker (void)
{	if  ( disp_magic )  RmvTime ( (QElem*)&tickerblock );
	disp_magic = false;
}



// *************************************************************************
// 								GLOBAL DATA
// *************************************************************************

Str63		specciWindowName;

long		frames_done		= 0;
long		frames			= 0;
int			hurry_up		= false;		// speed up, anyhow...

#if include_debugger
short			Run_Z80_debug();
short			Z80_debug();
extern			z80	ref_zreg;				// registers of reference engine
extern	Char	*REF_CORE;					// core of reference engine
Boolean	trace_mode			= false;		// debugger only
#endif

Boolean	use_better_z80		= false;		// use Z80()
Boolean	use_best_z80		= true;			// use Z80_PPC()

Boolean	run_unlimited		= false;		// no speed limit
Boolean	run_exact			= true;			// exact Specci speed
Boolean	run_highres			= false;		// highres screen update
Boolean	writable_rom		= false;

Boolean	z80_running			= false;		// z80 cpu: running or halted

int		z80_hardware		= zx48;			// currently emulated hardware
Char	zx128pagectrl		= 0;			// last byte sent to ZX 128's paging control port
long	line_cycle_adjust	= 0;			// adjust for video-page accesses in hires mode




// **************************************************************************
//					initializing & memory bank handling
// **************************************************************************


/* ----	Send byte to ZX 128's paging control port ----------------------------------------
		Bits 0-2:	Ram for top 16k:	0 is the usual bank;   7 contains some scratchpads used by the 128k editor.
		Bit 3:	Screen select:		0 => Use the normal screen in bank 5;	1 => Use the shadow screen in bank 7.
		Bit 4:	ROM for lower 16k:	0 => 128k editor;					1 => 48k BASIC
		Bit 5:	Lock. If set, disables all further paging.
*/
void ZX128PageCtrl(Char n)	
{
	if (z80_hardware!=zx128&&z80_hardware!=zxplus2) return;			// not a ZX 128
	if (zx128pagectrl&0x20) return;			// further paging disabled
	zx128pagectrl = n;
	//if (n==5||n==2) { GrafPtr saveport; GetPort(&saveport); ConChar('◊'); SetPort(saveport); }
	rpage[0] = (n&0x10) ? rom[1] : rom[0];		// ROM select
	if (writable_rom) wpage[0]=rpage[0]; // else wpage[0]=(Char*)norom;
	video_page = (n&0x08) ? ram[7] : ram[5];	// video RAM select
	rpage[3] = wpage[3] = ram[n&0x07]-0xc000;	// upper RAM page select
}


/* ----	Reset Specci --------------------------------------------------------
		Emulation mode is preserved
		reset Z80 cpu
		reset hardware: paging, video RAM, AY sound
		optionally reload ROMs
		set specci window name
		open specci window
*/
void HandleReset ( Boolean reload_roms )
{	OSErr err = ok;
	int i;

// Reset the Z80 cpu
	PC 		= SP	= 0;
	IFF1	=IFF2	= disabled;
	WUFF	= 0;				// irpt, nmi
	EXIT	= 0;				// watchdog
	IM		= 1;
	CYCLES	= 0;
//	TOTAL	= 0;	<-- causes major sound output problems: e_cnt, n_cnt and s_cnt[] contain next T cycle to change sound !!!  27.10.96 KIO !
	

// Reset paging & video RAM page
	zx128pagectrl = 0;
	for (i=4;i--;)	rpage[i] = wpage[i] = CORE;
	if (!writable_rom) wpage[0]=(Char*)norom; // else wpage[0]=rpage[0];
	
	video_page = RdPtr(0x4000);
	line_cycle_adjust = 0;

// Reset AY sound chip
	Set_AY_Register(7,0xff);		// disable all AY sound

// reload ROMs & set window name
	if (z80_hardware==zx48)
	{
		if (reload_roms) err = LoadResROM ( 48, rom[0] );
		SetSpecciWindowName ( "\pZX Spectrum 48k" );
	}
	else if (z80_hardware==zx128)
	{	if (reload_roms) err = LoadResROM ( 1280, rom[0] ) + LoadResROM ( 1281, rom[1] );
		SetSpecciWindowName ( "\pZX Spectrum 128k" );
	}
	else if (z80_hardware==zxplus2)
	{	if (reload_roms) err = LoadResROM ( 1282, rom[0] ) + LoadResROM ( 1283, rom[1] );
		SetSpecciWindowName ( "\pZX Spectrum +2" );
	}
	else do_abort("\pillegal model number","\pHandleReset()",0);

	StartStop(err==ok);
	NewSpecciWindow(scrnsize);
}


/* ----	Initialialize the z80 engine --------------------------------------------
		Initialize and reset Z80 engine (both, 68k and ppc version)
		Allocate Z80 memory and initialize page pointers
*/
void InitZ80 ( int hardware )
{	int i, roms, rams;
	OSErr err=noErr;
	Char *p;

	keyboard.quad.a = keyboard.quad.b = keyboard.quad.c = -1;	// clear keyboard matrix

	switch (hardware)
	{
	case zx16:		roms=1; rams=3; break;
	case zx48:		roms=1; rams=3; break;
	case zx128:		roms=2; rams=8; break;
	case zxplus2:	roms=2; rams=8; break;
	default:	do_abort("\pInitZ80():","\pillegal model number: ",PascalStr(NumStr(hardware)));
	}
	z80_hardware = hardware;

// allocate core and initialize z80 engine
	Init_Z80 (roms,rams );
	if (writable_rom) wpage[0]=rpage[0]; else wpage[0]=(Char*)norom;

// initialize rom[] and ram[]
	switch (z80_hardware)
	{
	case zx48:
		rom[0] = CORE;				// CORE+0x0000	= rom[0]	= rpage[0]
		ram[1] = CORE+0x04000;		// CORE+0x4000	= ram[1]	= rpage[1]
		ram[2] = CORE+0x08000;		// CORE+0x8000	= ram[2]	= rpage[2]
		ram[3] = CORE+0x0c000;		// CORE+0xC000	= ram[3] 	= rpage[3]
		break;
	case zx128:
	case zxplus2:
		memset(ram,0,sizeof(Char*)*maxrampages);	// kio 08.jan.2000
		rom[0] = CORE;				// CORE+0x0000	= rom[0]	= rpage[0]
		ram[5] = CORE+0x04000;		// CORE+0x4000	= ram[5]	= rpage[1]
		ram[2] = CORE+0x08000;		// CORE+0x8000	= ram[2]	= rpage[2]
		ram[0] = CORE+0x0c000;		// CORE+0xC000	= ram[0] 	= rpage[3]
		p=CORE;
		rom[1] = p-=pagesize;
		for (i=0;i<rams;i++) if (!ram[i]) { /*PrintChr('•');*/ ram[i] = p-=pagesize; }
		break;
	}

	HandleReset(true);
}


// **************************************************************************
//						running the virtual ZX Spectrum 
// **************************************************************************


// ----	Cycle counter reached 0 ------------------------------------------
//		if run_exact || run_unlimited trigger EXIT
//		if run_highres copy scanline to buffer and trigger EXIT at start of a new frame
//		••• detection and handling of video page switching missing !!! ••• ???

void Do_Cycles()
{
	if (run_highres)		// special handling in highres mode
	{	static struct	c32 *nptr,*cptr,*nattr,*cattr;
		static Char		*my_video_page;
		static int		scanline = 0;
		if (scanline-- > 0)		// 192 .. 1
		{	CYCLES	+= line_cycles - line_cycle_adjust;
			TOTAL	+= line_cycles;
			if (video_page!=my_video_page)
			{	cptr = (struct c32*)((Char*)cptr+(video_page-my_video_page));
				cattr = (struct c32*)((Char*)cattr+(video_page-my_video_page));
				my_video_page = video_page;
			}
			*nattr++	= *cattr;
			*nptr	= *cptr;
			if (scanline&0x07)	{ nptr+=8; cptr+=8; return; }
			cattr++; 
			if (scanline&0x38)	{ nptr-=8*7-1; cptr-=8*7-1; return; }
			if (scanline)		{ nptr++; cptr++; return; }
			// border below screen:
			CYCLES	+= frame_cycles-(192+above_scrn)*line_cycles;	
			TOTAL	+= frame_cycles-(192+above_scrn)*line_cycles;
			return;
		};
		// interrupt&new frame: border above screen:
		
		nptr	= (struct c32*)new_pixels;
		cptr 	= (struct c32*)video_page;
		nattr	= (struct c32*)new_attr;
		cattr	= (struct c32*)cptr+192;
		my_video_page = video_page;
		scanline = 192;
	}
	else if (run_unlimited && frames_done-frames>=0) 
	{
		CYCLES += line_cycles*8;
		TOTAL  += line_cycles*8;
		return;
	}

	EXIT 	= true;	// new frame  =>  force Z80() to return
}


// ----	Perform regular actions --------------------------------------------------
static void Perform_regular_actions(void)
{	static int cnt=50;

// play sound (if sound on)
// may set hurry_up or may wait to synchronize timing
	PlaySample();		

// refresh screen and poll events
// ensure at least 1 refresh and poll per second
	if (!hurry_up || --cnt==0)
	{
		cnt=50;
		if (mbar_present) PollEvents(everyEvent,0,nil);
		RefreshSpecciWindow();
	}
	brdrlogptr = brdrlog;			// only required for hires display


// prepare z80() for next frame
	CYCLES		+= run_highres ? line_cycles*above_scrn : frame_cycles;
	TOTAL		+= run_highres ? line_cycles*above_scrn : frame_cycles;
	EXIT		= false;
	WUFF 		|= 1;


// scan keyboard
	UpdateKeyboard();

// load / save movie frame
#if include_movie
	DoMovieFrame();
#endif
}


// ----	RUN ZX Spectrum -----------------------------------------------------
void RunZ80(void)
{	short n;
	
	SetSpecciWindowName ( specciWindowName );

	EnableItem ( filemenu,iSuspend);
	DisableItem ( filemenu,iResume);	

	EnableItem ( debugmenu, iTriggerIrpt );
	EnableItem ( debugmenu, iTriggerNMI );
	EnableItem ( debugmenu, iTriggerReset );

	if (powerMgr) DisableIdle();
	
	LMSetSysEvtMask ( LMGetSysEvtMask() | keyUpMask );		// allow keyUp Events	
	
	frames_done = frames;

	ResumeSound();

	while ( !done && z80_running )
	{	

	#if !GENERATINGPOWERPC
		if (use_better_z80)  n = Z80(); else
	#endif
		n = Z80_PPC(); 

		switch ( n )
		{	
		case rst0_instr:
			Handle_RST0();
			continue;
		
		case halt_instr:
			if (IFF1==disabled) 
			{
				PrintAlert(CStr("••••• HALT instruction executed with interrupts disabled"));
				goto p;
			}
			while (!EXIT) Do_Cycles();				// if highresmode collect remaining scan lines
			if (CYCLES>0) CYCLES = 0;				// wait for interrupt -> no T cycles left
		//	resume with watchdog_irpt
	
		case watchdog_irpt:
			Perform_regular_actions();
			continue;
	
		case irpt_error:
			PrintAlert(CStr("••••• Internal error: Interrupt failure"));
			goto p;

		case nimp_instr:
			PrintAlert(CStr("••••• Not implemented instruction:"));
			goto p;

		case ill_instr:
			PrintAlert(CStr("••••• Illegal instruction trapped:"));
			goto p;

		case weird_instr:
			PrintAlert(CStr("••••• Weird instruction trapped:"));
			goto p;

		default:
			PrintNumberAlert ( CStr("••••• Internal error: Z80 bad result: "), n );

		p:	PrintHex2(Peek(PC-3));
			PrintHex2(Peek(PC-2));
			PrintHex2(Peek(PC-1));
			PrintStr(CStr("<-- "));
			PrintHex2(Peek(PC));
			PrintHex2(Peek(PC+1));
			PrintHex2(Peek(PC+2));
			PrintHex2(Peek(PC+3));
			PrintStr(CStr("at:"));
			PrintHex4(PC);
			PrintNL();
			HandleFileMenu(iSuspend);
			break;
		};

	};


// ----	Z80 suspended -----------------------------------------------------------
	
	SuspendSound(); 

	SetSpecciWindowName ( specciWindowName );

	if (zxwindow)
	{	Rect box;
		SetRect (&box, 0, 0, 31, 11);		// remove speed display ((required only if running unlimited))
		SetPort(zxwindow);
		InvalRect(&box);
	};

	DisableItem(filemenu,iSuspend);
	EnableItem(filemenu,iResume);
	
	DisableItem ( debugmenu, iTriggerIrpt );
	DisableItem ( debugmenu, iTriggerNMI );
	DisableItem ( debugmenu, iTriggerReset );

	if (powerMgr) EnableIdle();
	
	LMSetSysEvtMask ( LMGetSysEvtMask() & ~keyUpMask );
	
	while ( !done && !z80_running )	
	{	PollEvents ( everyEvent, 10, nil );
		RefreshSpecciWindow();
	};
}


